"""
Parse a dictionary structure and return an immutable structure that
contain a validated configuration.

This module contains two sets of classes.

Config objects: These are immutable structures (namedtuples, FrozenDict) that
contain the configuration data. The top level structure (ConfigTron) is returned
from valid_config().

Validator objects: These are responsible for validating a dictionary structure
and returning a valid immutable config object.
"""

from collections import namedtuple
from functools import partial
import itertools
import logging
import os
import re

import pytz
import yaml

from tron.config import ConfigError
from tron.config.schedule_parse import valid_schedule
from tron.utils.dicts import FrozenDict
from tron.core.action import CLEANUP_ACTION_NAME


log = logging.getLogger("tron.config")

YAML_TAG_RE = re.compile(r'!\w+\b')

def load_config(string_or_file):
    """Given a string or file object, load it with PyYAML and return an
    immutable, validated representation of the configuration it specifies.
    """
    # TODO: 0.5 remove this
    # Hackishly strip !Tags from the file so PyYAML doesn't try to make them
    # into Python objects
    if not isinstance(string_or_file, basestring):
        s_tags = string_or_file.read()
    else:
        s_tags = string_or_file
    s_notags = YAML_TAG_RE.sub('', s_tags)

    if len(s_tags) > len(s_notags):
        log.warn('Tron no longer uses !Tags to parse config files. Please'
                 ' remove them from yours.')

    # Load with YAML. safe_load() disables python classes
    config = yaml.safe_load(s_notags)

    return valid_config(config)


### SCHEMA DEFINITION ###

def config_object_factory(name, required=None, optional=None):
    """
    Creates a namedtuple which has two additional attributes:
        required_keys:
            all keys required to be set on this configuration object
        optional keys:
            optional keys for this configuration object

    The tuple is created from required + optional
    """
    required = required or []
    optional = optional or []
    config_class = namedtuple(name, required + optional)
    config_class.required_keys = required
    config_class.optional_keys = optional
    return config_class


TronConfig = config_object_factory(
    'TronConfig',
    optional=[
         'working_dir',         # str
         'output_stream_dir',   # str
         'command_context',     # FrozenDict of str
         'ssh_options',         # ConchOptions
         'notification_options',# NotificationOptions or None
         'time_zone',           # pytz time zone
         'nodes',               # FrozenDict of ConfigNode
         'node_pools',          # FrozenDict of ConfigNodePool
         'jobs',                # FrozenDict of ConfigJob
         'services'             # FrozenDict of ConfigService
    ])


NotificationOptions = config_object_factory(
    'NotificationOptions',
    [
        'smtp_host',            # str
        'notification_addr',    # str
    ])


ConfigSSHOptions = config_object_factory(
    'ConfigSSHOptions',
    optional=[
        'agent',                # bool
        'identities',           # list of str
    ])


ConfigNode = config_object_factory('ConfigNode', ['hostname'], ['name'])


ConfigNodePool = config_object_factory('ConfigNodePool', ['nodes'], ['name'])


ConfigJob = config_object_factory(
    'ConfigJob',
    [
        'name',                 # str
        'node',                 # str
        'schedule',             # Config*Scheduler
        'actions',              # FrozenDict of ConfigAction
    ],[
        'queueing',             # bool
        'run_limit',            # int
        'all_nodes',            # bool
        'cleanup_action',       # ConfigAction
        'enabled',              # bool
    ])


ConfigAction = config_object_factory(
    'ConfigAction',
    [
        'name',                 # str
        'command',              # str
    ],[
        'requires',             # tuple of str
        'node',                 # str
    ])

ConfigCleanupAction = config_object_factory(
    'ConfigCleanupAction',
    [
        'command',              # str
    ],[
        'requires',             # tuple of str
        'name',                 # str
        'node',                 # str
    ])


ConfigService = config_object_factory(
    'ConfigService',
    [
        'name',                 # str
        'node',                 # str
        'pid_file',             # str
        'command',              # str
        'monitor_interval',     # float
    ],[
        'restart_interval',     # float
        'count',                # int
    ])


### VALIDATION ###

class UniqueNameDict(dict):
    """A dict like object that throws a ConfigError if a key exists and a set
    is called to change the value of that key.
     *fmt_string* will be interpolated with (key,)
    """
    def __init__(self, fmt_string, **kwargs):
        super(dict, self).__init__(**kwargs)
        self.fmt_string = fmt_string

    def __setitem__(self, key, value):
        if key in self:
            raise ConfigError(self.fmt_string % key)
        super(UniqueNameDict, self).__setitem__(key, value)


def type_converter(convert, error_fmt):
    def f(path, value, optional=False):
        """Convert *value* at config path *path* into something else via the
        *convert* function, raising ConfigError if *convert* raises a
        TypeError. *error_fmt* will be interpolated with (path, value).
        If optional is True, None values will be returned without converting.
        """
        if value is None and optional:
            return None
        try:
            return convert(value)
        except TypeError:
            raise ConfigError(error_fmt % (path, value))
    return f

valid_float = type_converter(float, 'Value at %s is not a number: %s')

valid_int = type_converter(int, 'Value at %s is not an integer: %s')


def type_validator(validator, error_fmt):
    def f(path, value, optional=False):
        """If *validator* does not return True for *value*, raise ConfigError
        If optional is True, None values will be returned without validation.
        """
        if value is None and optional:
            return None
        if not validator(value):
            raise ConfigError(error_fmt % (path, value))
        return value
    return f

valid_str = type_validator(
    lambda s: isinstance(s, basestring),
    'Value at %s is not a string: %s')

IDENTIFIER_RE = re.compile(r'^[A-Za-z_]\w*$')
valid_identifier = type_validator(
    lambda s: IDENTIFIER_RE.match(s),
    'Identifier at %s is not a valid identifier')

valid_list = type_validator(
    lambda s: isinstance(s, list), 'Value at %s is not a list: %s')

valid_populated_list = type_validator(
    bool, 'Value at %s is not a list with items: %s')

valid_dict = type_validator(
    lambda s: isinstance(s, dict), 'Value at %s is not a dictionary: %s')

valid_bool = type_validator(
    lambda s: isinstance(s, bool), 'Value at %s is not a boolean: %s')


class Validator(object):
    """Base class for validating a collection and creating a mutable
    collection from the source.
    """
    config_class =              None
    defaults =                  {}
    validators =                {}
    optional =                  False

    def validate(self, in_dict):
        if self.optional and in_dict is None:
            return None

        if in_dict is None:
            raise ConfigError("A %s is required." % self.type_name)

        shortcut_value = self.do_shortcut(in_dict)
        if shortcut_value:
            return shortcut_value

        in_dict = self.cast(in_dict)
        self.validate_required_keys(in_dict)
        self.validate_extra_keys(in_dict)
        output_dict = self.build_dict(in_dict)
        self.set_defaults(output_dict)
        return self.config_class(**output_dict)

    __call__ = validate

    @property
    def type_name(self):
        """Return a string that represents the config_class being validated.
        This name is used for error messages, so we strip off the word
        Config so the name better matches what the user sees in the config.
        """
        return self.config_class.__name__.replace("Config", "")

    def do_shortcut(self, in_dict):
        """Override if your validator can skip most of the validation by
        checking this condition.  If this returns a truthy value, the
        validation will end immediately and return that value.
        """
        pass

    def cast(self, in_dict):
        """If your validator accepts input in different formations, override
        this method to cast your input into a common format.
        """
        return in_dict

    def validate_required_keys(self, in_dict):
        """Check that all required keys are present."""
        missing_keys = set(self.config_class.required_keys) - set(in_dict)
        if not missing_keys:
            return

        keys = self.config_class.required_keys + self.config_class.optional_keys
        missing_key_str = ', '.join(missing_keys)
        if 'name' in keys and 'name' in in_dict:
            raise ConfigError("%s %s is missing options: %s" % (
                    self.type_name, in_dict['name'], missing_key_str)
            )

        raise ConfigError("Nameless %s is missing options: %s" % (
                self.type_name, missing_key_str)
        )

    def validate_extra_keys(self, in_dict):
        """Check that no unexpected keys are present."""
        all_keys = self.config_class.required_keys + self.config_class.optional_keys
        extra_keys = set(in_dict) - set(all_keys)
        if not extra_keys:
            return

        raise ConfigError("Unknown options in %s %s: %s" % (
                self.type_name,
                in_dict.get('name', ''),
                ', '.join(extra_keys)
        ))

    def set_defaults(self, output_dict):
        """Set any default values for any optional values that were not
        specified.
        """
        for key, value in self.defaults.iteritems():
            if key not in output_dict:
                output_dict[key] = value

    def post_validation(self, valid_input):
        """Perform additional validation."""
        pass

    def build_dict(self, input):
        """Override this to validate each value in the input."""
        valid_input = {}
        for key, value in input.iteritems():
            if key in self.validators:
                valid_input[key] = self.validators[key](value)
            else:
                valid_input[key] = value
        self.post_validation(valid_input)
        return valid_input


class ValidatorWithNamedPath(Validator):
    """A validator that expects a name to use for validation failure messages
    and calls a post_validation() method after building the dict."""

    def post_validation(self, valid_input, path_name):
        pass

    def build_dict(self, input):
        path_name = '%s.%s' % (self.type_name, input.get('name'))
        valid_input = {}
        for key, value in input.iteritems():
            if key in self.validators:
                valid_input[key] = self.validators[key](path_name, value)
            else:
                valid_input[key] = value
        self.post_validation(valid_input, path_name)
        return valid_input


# TODO: remove in 0.5
def valid_working_dir(wd):
    """Given a working directory or None, return a valid working directory.
    If wd=None the mcp will attempt to use a default.
    """
    log.warning("working_dir is no longer supported. "
        "Use output_stream_dir to set the directory for stdout/stderr files.")


def valid_output_stream_dir(output_dir):
    """Returns a valid string for the output directory, or raises ConfigError
    if the output_dir is not valid.
    """
    output_dir = valid_str('output_stream_dir', output_dir, optional=True)
    if not output_dir:
        return

    if not os.path.isdir(output_dir):
        msg = "output_stream_dir '%s' is not a directory"
        raise ConfigError(msg % output_dir)

    if not os.access(output_dir, os.W_OK):
        raise ConfigError("output_stream_dir '%s' is not writable" % output_dir)

    return output_dir


def valid_command_context(context):
    # context can be any dict.
    return FrozenDict(**valid_dict('command_context', context or {}))


def valid_time_zone(tz):
    if tz is None:
        return None
    try:
        return pytz.timezone(valid_str('time_zone', tz))
    except pytz.exceptions.UnknownTimeZoneError:
        raise ConfigError('%s is not a valid time zone' % tz)


class ValidateSSHOptions(Validator):
    """Validate SSH options."""
    config_class =              ConfigSSHOptions
    optional =                  True
    defaults = {
        'agent':                False,
        'identities':           ()
    }
    validators = {
        'agent':                partial(valid_bool, 'ssh_options.agent'),
        'identities':           partial(valid_list, 'ssh_options.identities')
    }

valid_ssh_options = ValidateSSHOptions()


class ValidateNotificationOptions(Validator):
    """Validate notification options."""
    config_class =              NotificationOptions
    optional =                  True

valid_notification_options = ValidateNotificationOptions()


# TODO: remove in 0.5
def normalize_node(node):
    """Given a node value from a config, determine if it's the node's name, the
    node's value, or actually a node pool. The former case is the "new style"
    of identifiers and the latter case is the "old style" of anchors and
    aliases (*/&). Return the name.
    """
    if node is None:
        return None
    if isinstance(node, basestring):
        return node
    # probably a reference back to an already validated node
    try:
        return valid_node(node).name
    except ConfigError:
        return valid_node_pool(node).name

class ValidateNode(Validator):
    config_class =              ConfigNode

    def do_shortcut(self, node):
        """Nodes can be specified with just a hostname string."""
        if isinstance(node, basestring):
            return ConfigNode(hostname=node, name=node)

    def set_defaults(self, output_dict):
        output_dict.setdefault('name', output_dict['hostname'])

valid_node = ValidateNode()


class ValidateNodePool(Validator):
    config_class =              ConfigNodePool

    def cast(self, node_pool):
        if isinstance(node_pool, list):
            node_pool = dict(nodes=node_pool)
        if 'nodes' in node_pool:
            node_pool['nodes'] = [
                normalize_node(node) for node in node_pool['nodes']
            ]
        return node_pool

    def set_defaults(self, node_pool):
        node_pool.setdefault('name', '_'.join(node_pool['nodes']))

valid_node_pool = ValidateNodePool()


class ValidateAction(ValidatorWithNamedPath):
    """Validate an action."""
    config_class =              ConfigAction
    defaults = {
        'node':                 None
    }
    validators = {
        'name':                 valid_identifier,
        'command':              valid_str,
        'node':                 lambda _, v: normalize_node(v)
    }

    def post_validation(self, action, path_name):
        # check name
        if action['name'] == CLEANUP_ACTION_NAME:
            raise ConfigError("Bad action name at %s: %s" %
                              (path_name, action['name']))

        requires = []

        # accept a string, pointer, or list
        old_requires = action.get('requires', [])

        # string identifier
        if isinstance(old_requires, basestring):
            log.warn("Require without a list is deprecated. "
                "You should update requires for %s %s" %
                (path_name, action['name']))
            old_requires = [old_requires]

        # pointer
        if isinstance(old_requires, dict):
            old_requires = [old_requires['name']]

        old_requires = valid_list(path_name, old_requires)

        for r in old_requires:
            if not isinstance(r, basestring):
                # old style, alias
                r = r['name']

            requires.append(r)
            if r == CLEANUP_ACTION_NAME:
                raise ConfigError('Actions cannot depend on the cleanup action.'
                                  ' (%s)' % path_name)

        action['requires'] = tuple(requires)

valid_action = ValidateAction()


class ValidateCleanupAction(ValidatorWithNamedPath):
    config_class =              ConfigCleanupAction
    defaults = {
        'node':                 None,
        'name':                 CLEANUP_ACTION_NAME,
    }
    validators = {
        'name':                 valid_identifier,
        'command':              valid_str,
        'node':                 lambda _, v: normalize_node(v)
    }

    def post_validation(self, action, path_name):
        if (
            'name' in action
            and action['name'] not in (None, CLEANUP_ACTION_NAME)
        ):
            raise ConfigError("Cleanup actions cannot have custom names (you"
                              " wanted %s.%s)" % (path_name, action['name']))

        if 'requires' in action:
            raise ConfigError("Cleanup action %s can not have requires." % path_name)

        action['requires'] = tuple()

valid_cleanup_action = ValidateCleanupAction()


class ValidateJob(ValidatorWithNamedPath):
    """Validate jobs."""
    config_class =              ConfigJob
    defaults = {
        'run_limit':            50,
        'all_nodes':            False,
        'cleanup_action':       None,
        'enabled':              True,
        'queueing':             True
    }

    validators = {
        'name':                 valid_identifier,
        'schedule':             valid_schedule,
        'run_limit':            valid_int,
        'all_nodes':            valid_bool,
        'actions':              valid_populated_list,
        'cleanup_action':       lambda _, v: valid_cleanup_action(v),
        'node':                 lambda _, v: normalize_node(v),
        'queueing':             valid_bool,
        'enabled':              valid_bool,
    }

    def _validate_dependencies(self, job, actions,
        base_action, current_action=None, stack=None):
        """Check for circular or misspelled dependencies."""
        stack = stack or []
        current_action = current_action or base_action

        stack.append(current_action.name)
        for dep in current_action.requires:
            if dep == base_action.name and len(stack) > 0:
                raise ConfigError(
                    'Circular dependency in job.%s: %s' % (
                    job['name'], ' -> '.join(stack)))
            if dep not in actions:
                raise ConfigError(
                    'Action jobs.%s.%s has a dependency "%s"'
                    ' that is not in the same job!' %
                    (job['name'], current_action.name, dep))
            self._validate_dependencies(
                job, actions, base_action, actions[dep], stack)

        stack.pop()

    def post_validation(self, job, path):
        """Validate actions for the job."""
        actions = UniqueNameDict('Action name %%s on job %s used twice' % job['name'])
        for action in job['actions'] or []:
            try:
                final_action = valid_action(action)
            except ConfigError, e:
                raise ConfigError("Invalid action config for %s: %s" % (path, e))

            if not (final_action.node or job['node']):
                raise ConfigError(
                    '%s has no node configured for %s' % (path, final_action.name))
            actions[final_action.name] = final_action

        for action in actions.values():
            self._validate_dependencies(job, actions, action)

        job['actions'] = FrozenDict(**actions)

valid_job = ValidateJob()


class ValidateService(ValidatorWithNamedPath):
    """Validate a services configuration."""
    config_class =              ConfigService
    defaults = {
        'count':                1,
        'restart_interval':     None
    }

    validators = {
        'name':                 valid_identifier,
        'pid_file':             valid_str,
        'command':              valid_str,
        'monitor_interval':     valid_int,
        'count':                valid_int,
        'node':                 lambda _, v: normalize_node(v),
        'restart_interval':     valid_int,
    }

valid_service = ValidateService()


class ValidateConfig(Validator):
    """Given a parsed config file (should be only basic literals and
    containers), return an immutable, fully populated series of namedtuples and
    FrozenDicts with all defaults filled in, all valid values, and no unused
    values. Throws a ConfigError if any part of the input dict is invalid.
    """
    config_class =              TronConfig
    defaults = {
        'working_dir':          None,
        'output_stream_dir':    None,
        'command_context':      None,
        'ssh_options':          valid_ssh_options({}),
        'notification_options': None,
        'time_zone':            None,
        'nodes':                (dict(name='localhost', hostname='localhost'),),
        'node_pools':           (),
        'jobs':                 (),
        'services':             (),
    }
    validators = {
        'output_stream_dir':    valid_output_stream_dir,
        'working_dir':          valid_working_dir,
        'command_context':      valid_command_context,
        'ssh_options':          valid_ssh_options,
        'notification_options': valid_notification_options,
        'time_zone':            valid_time_zone,
    }
    optional = False

    def post_validation(self, config):
        """Validate jobs, nodes, and services."""

        node_names = UniqueNameDict('Node and NodePool names must be unique %s')
        job_service_names = UniqueNameDict('Job and Service names must be unique %s')

        # We need to set this default here until 0.5, see comment below
        config.setdefault('node_pools', [])
        # 'nodes' may contain nodes or node pools (for now). Internally split them
        # into 'nodes' and 'node_pools'. Process the nodes.
        nodes = config.pop('nodes', None) or []
        config['nodes'] = []
        for node in nodes:
            # TODO: 0.5 remove
            # A node pool is characterized by being a list, or a dictionary that
            # contains a list of nodes
            if isinstance(node, (list, dict)) and 'nodes' in node:
                log.warn('Node pools should be moved from "nodes" to "node_pools"'
                         ' before upgrading to Tron 0.5.')
                config['node_pools'].append(node)
            else:
                config['nodes'].append(node)

        def parse_sub_config(cname, valid, name_dict):
            target_dict = UniqueNameDict(
                "%s name %%s used twice" % cname.replace('_', ' '))
            for item in config.get(cname) or []:
                final = valid(item)
                target_dict[final.name] = final
                name_dict[final.name] = True
            config[cname] = FrozenDict(**target_dict)

        parse_sub_config('nodes',       valid_node,         node_names)
        parse_sub_config('node_pools',  valid_node_pool,    node_names)
        parse_sub_config('jobs',        valid_job ,         job_service_names)
        parse_sub_config('services',    valid_service,      job_service_names)

        self.validate_node_names(config, node_names)
        self.validate_node_pool_nodes(config)

    def validate_node_names(self, config, node_names):
        """Validate that any node/node_pool name that were used are configured
        as nodes/node_pools.
        """
        actions = itertools.chain.from_iterable(
            job.actions.values()
            for job in config['jobs'].values()
        )
        task_list = itertools.chain(
            config['jobs'].values(),
            config['services'].values(),
            actions
        )
        for task in task_list:
            if task.node and task.node not in node_names:
                raise ConfigError("Unknown node %s configured for %s %s" % (
                    task.node, task.__class__.__name__, task.name))

    def validate_node_pool_nodes(self, config):
        """Validate that each node in a node_pool is in fact a node, and not
        another pool.
        """
        # TODO: this can be cleaned up after 0.5
        for node_pool in config['node_pools'].itervalues():
            for node_name in node_pool.nodes:
                node = config['nodes'].get(node_name)
                if node:
                    continue
                raise ConfigError(
                    "NodePool %s contains another NodePool %s. " % (
                        node_pool.name,
                        node_name
                    )
                )


valid_config = ValidateConfig()



import heapq
from collections import deque, namedtuple
import logging
import weakref
import time

from tron.utils import timeutils
from tron.utils import observer

log = logging.getLogger(__name__)

# Event Levels INFO is for troubleshooting information. This may be verbose but
# shouldn't cause any monitors to make any decisions.
LEVEL_INFO = "INFO"

# OK indicates the entity is doing great and any monitors that considered the
# entity to be in non-ok state can reset itself.
LEVEL_OK = "OK"

# NOTICE indicates some troubling behavior, but not yet a complete failure. It
# would be appropriate to highlight this event, but don't go waking up the
# president just yet.
LEVEL_NOTICE = "NOTICE"

# CRITICAL indicates the entity has had a major failure. Call in the troops.
LEVEL_CRITICAL = "CRITICAL"

# To allow our levels to be ordered, we provide this list. Use .index(level) to
# be able to compare levels
ORDERED_LEVELS = [
    LEVEL_INFO,
    LEVEL_OK,
    LEVEL_NOTICE,
    LEVEL_CRITICAL,
]

EventType = namedtuple('EventType', ['level', 'name'])


class FixedLimitStore(object):
    """Simple data store that keeps a fixed number of elements based on their
    'category'. Also known as a circular buffer or ring buffer.
    """
    DEFAULT_LIMIT = 10

    def __init__(self, limits):
        self._limits = limits or dict()
        self._values = {}

    def _build_deque(self, category):
        limit = self._limits.get(category, self.DEFAULT_LIMIT)
        return deque(maxlen=limit)

    def append(self, category, item):
        if category not in self._values:
            self._values[category] = self._build_deque(category)
        self._values[category].append((time.time(), item))

    def _build_iter(self, categories):
        event_groups = [self._values.get(cat, []) for cat in categories]
        return (val for _, val in heapq.merge(*event_groups))

    def __iter__(self):
        return self._build_iter(self._values.keys())

    def list(self, categories):
        return list(self._build_iter(categories))


# TODO: src should just be the entity, not an EventRecorder
class Event(object):
    """Data object for storing details of an event."""
    __slots__ = ('_src', 'time', 'level', 'name', 'data')

    def __init__(self, src, level, name, **data):
        self._src = weakref.ref(src)
        self.time = timeutils.current_time()
        self.level = level
        self.name = name
        self.data = data

    @property
    def entity(self):
        src = self._src()
        return src.entity if src else None


class EventRecorder(observer.Observer):
    """Record events in a tree of listeners. Tron uses this class by having
    one process-wide event recorder (in the MCP) with each job and service
    having a child recorder, and each job run having a child recorder of the
    job.

    Events are propagated up the chain if they are of high enough severity.
    """

    def __init__(self, entity, parent=None, limits=None):
        self._store = FixedLimitStore(limits)
        self._parent = None
        self._entity = weakref.ref(entity)
        self.watch(entity)

        if parent:
            self.set_parent(parent)

    def set_parent(self, parent):
        self._parent = weakref.ref(parent)

    def _get_entity(self):
        return self._entity()

    def _set_entity(self, entity):
        self._entity = weakref.ref(entity)

    entity = property(_get_entity, _set_entity)

    def record(self, event):
        self._store.append(event.level, event)

        # Propagate if we have a parent set (and the level is high enough to
        # care)
        if (self._parent and
            ORDERED_LEVELS.index(event.level) >
                ORDERED_LEVELS.index(LEVEL_INFO)):
            self._parent().record(event)

    def emit_info(self, name, **data):
        self.record(Event(self, LEVEL_INFO, name, **data))

    def emit_ok(self, name, **data):
        self.record(Event(self, LEVEL_OK, name, **data))

    def emit_notice(self, name, **data):
        self.record(Event(self, LEVEL_NOTICE, name, **data))

    def emit_critical(self, name, **data):
        self.record(Event(self, LEVEL_CRITICAL, name, **data))

    def list(self, min_level=None):
        """Return a list of Events. If min_level is set, then only events
        with equal or higher priority will be returned.
        """
        levels = ORDERED_LEVELS
        if min_level is not None:
            levels = levels[levels.index(min_level):]
        return self._store.list(levels)

    # TODO: once Event is fixed, accept Event objects as well
    def handler(self, observable, event):
        """Watch for events and create and store Event objects."""
        if not isinstance(event, EventType):
            return

        self.record(Event(self, event.level, event.name))


class EventManager(object):
    """Create and store EventRecorder objects for observable objects.
    This class is a singleton.
    """

    _instance = None

    def __init__(self):
        if self._instance is not None:
            raise ValueError(
                    "EventManager already instantiated. Use get_instance().")

        self.recorders = {}

    @classmethod
    def get_instance(cls):
        if cls._instance is None:
            cls._instance = cls()
        return cls._instance

    def _build_key(self, observable):
        """Create a unique key for this observable object.
        EventManager makes the assumption that objects str values will be
        unique.
        """
        return str(observable)

    def add(self, observable, parent=None):
        """Create an EventRecorder for the observable and store it."""
        key = self._build_key(observable)
        if key in self.recorders:
            log.warn("%s is already being managed." % observable)

        parent_recorder = None
        if parent:
            parent_recorder = self.get(parent)
            if not parent_recorder:
                log.warn("Parent %s of %s is not being managed." % (
                        parent, observable))

        event_recorder = EventRecorder(observable, parent_recorder)
        self.recorders[key] = event_recorder
        return event_recorder

    def get(self, observable):
        """Return an EventRecorder given an observable."""
        key = self._build_key(observable)
        return self.recorders.get(key, None)

    def clear(self):
        self.recorders.clear()

    def remove(self, observable):
        key = self._build_key(observable)
        if key in self.recorders:
            del self.recorders[key]



"""
Web Services Interface used by command-line clients and web frontend to
view current state, event history and send commands to trond.
"""

import datetime
import logging
import urllib

try:
    import simplejson as json
    assert json # pyflakes
except ImportError:
    import json

from twisted.web import http, resource, server

from tron import service, event
from tron.api import adapter
from tron.core import actionrun
from tron.api import requestargs


log = logging.getLogger(__name__)


class JSONEncoder(json.JSONEncoder):
    """Custom JSON for certain objects"""

    def default(self, o):
        if isinstance(o, datetime.datetime):
            return o.strftime("%Y-%m-%d %H:%M:%S")

        if isinstance(o, datetime.date):
            return o.isoformat()

        return super(JSONEncoder, self).default(o)


def respond(request, response_dict, code=http.OK, headers=None):
    """Helper to generate a json response"""
    request.setResponseCode(code)
    request.setHeader('content-type', 'text/json')
    if headers:
        for key, val in headers.iteritems():
            request.setHeader(key, val)
    if response_dict:
        return json.dumps(response_dict, cls=JSONEncoder)
    return ""


class ActionRunResource(resource.Resource):

    isLeaf = True

    def __init__(self, job_run, action_name):
        resource.Resource.__init__(self)
        self._job_run           = job_run
        self._action_name       = action_name

    def render_GET(self, request):
        num_lines = requestargs.get_integer(request, 'num_lines')
        run_adapter = adapter.ActionRunAdapter(
                self._job_run, self._action_name, num_lines)
        return respond(request, run_adapter.get_repr())

    def render_POST(self, request):
        cmd = requestargs.get_string(request, 'command')
        log.info("Handling '%s' request for action run %s",
                 cmd, self._job_run.id, self._action_name)

        if cmd not in ('start', 'success', 'cancel', 'fail', 'skip'):
            log.warning("Unknown request command %s", cmd)
            return respond(request, None, code=http.NOT_IMPLEMENTED)

        action_run = self._job_run.action_runs[self._action_name]
        try:
            resp = getattr(action_run, cmd)()
        except actionrun.Error:
            resp = None

        if not resp:
            msg = "Failed to %s action run %s." % (cmd, action_run)
        else:
            msg = "Action run now in state %s" % action_run.state.short_name
        return respond(request, {'result': msg})


class JobRunResource(resource.Resource):

    isLeaf = False

    def __init__(self, run, master_control):
        resource.Resource.__init__(self)
        self._run = run
        self._master_control = master_control

    def getChild(self, act_name, request):
        if act_name == '':
            return self
        if act_name == '_events':
            return EventResource(self._run)
        if act_name in self._run.action_runs:
            return ActionRunResource(self._run, act_name)

        return resource.NoResource("Cannot find action '%s' for job run '%s'" %
                                   (act_name, self._run.id))

    def render_GET(self, request):
        run_adapter = adapter.JobRunAdapter(self._run, include_action_runs=True)
        return respond(request, run_adapter.get_repr())

    def render_POST(self, request):
        cmd = requestargs.get_string(request, 'command')
        log.info("Handling '%s' request for job run %s", cmd, self._run.id)

        if cmd not in ['start', 'restart', 'success', 'fail', 'cancel']:
            log.warning("Unknown request command %s", cmd)
            return respond(request, None, code=http.NOT_IMPLEMENTED)

        getattr(self, '_%s' % cmd)()
        return respond(request, {'result': "Job run now in state %s" %
                                 self._run.state.short_name})

    def _restart(self):
        log.info("Resetting all action runs to scheduled state")
        job_name = self._run.job_name
        job_sched = self._master_control.jobs[job_name]
        job_sched.manual_start(self._run.run_time)

    def _start(self):
        if self._run.start():
            log.info("Starting job run %s", self._run.id)
        else:
            log.warning("Failed to start job run %s" % self._run)

    def _success(self):
        if self._run.success():
            log.info("Marking job run %s for success", self._run.id)
        else:
            log.warning("Request to mark job run %s succeed when it has"
                        " already", self._run.id)

    def _cancel(self):
        if self._run.cancel():
            log.info("Cancelling job %s", self._run.id)
            self._run.cancel()
        else:
            log.warning("Request to cancel job run %s when it's already"
                        " cancelled", self._run.id)

    def _fail(self):
        if self._run.fail():
            log.info("Marking job run %s as failed", self._run.id)
        else:
            log.warning("Request to fail job run %s when it's already running"
                        " or done", self._run.id)


class JobResource(resource.Resource):
    """A resource that describes a particular job"""

    isLeaf = False

    def __init__(self, job_sched, master_control):
        self._job_sched = job_sched
        self._master_control = master_control
        resource.Resource.__init__(self)

    def getChild(self, run_id, request):
        job = self._job_sched.job
        if run_id == '':
            return self
        if run_id == '_events':
            return EventResource(self._job_sched.job)

        run_id = run_id.upper()
        if run_id == 'HEAD':
            run = job.runs.get_newest()
        elif run_id.isdigit():
            run = job.runs.get_run_by_num(int(run_id))
        else:
            run = job.runs.get_run_by_state_short_name(run_id)

        if run:
            return JobRunResource(run, self._master_control)
        return resource.NoResource(
                "Cannot find run number '%s' for job '%s'" % (run_id, job.name))

    def render_GET(self, request):
        include_action_runs = requestargs.get_bool(request, 'include_action_runs')
        job_adapter = adapter.JobAdapter(
                self._job_sched.job, True, include_action_runs)
        return respond(request, job_adapter.get_repr())

    def render_POST(self, request):
        cmd = requestargs.get_string(request, 'command')
        log.info("Handling '%s' request for job run %s",
                cmd, self._job_sched.job.name)

        if cmd == 'enable':
            self._job_sched.enable()
            msg = "Job %s is enabled" % self._job_sched.job.name

        elif cmd == 'disable':
            self._job_sched.disable()
            msg = "Job %s is disabled" % self._job_sched.job.name

        elif cmd == 'start':
            run_time = requestargs.get_datetime(request, 'run_time')
            runs = self._job_sched.manual_start(run_time=run_time)
            msg = "New Job Runs %s created" % ",".join([r.id for r in runs])

        else:
            return respond(request, None, code=http.NOT_IMPLEMENTED)

        return respond(request, {'result': msg})


class JobsResource(resource.Resource):
    """Resource for all our daemon's jobs"""

    def __init__(self, master_control):
        self._master_control = master_control
        resource.Resource.__init__(self)

    def getChild(self, name, request):
        if name == '':
            return self

        job_sched = self._master_control.jobs.get(name)
        if job_sched is None:
            return resource.NoResource("Cannot find job '%s'" % name)

        return JobResource(job_sched, self._master_control)

    def get_data(self, include_job_run=False, include_action_runs=False):
        job_adapter = adapter.JobAdapter
        return [
            job_adapter(job.job, include_job_run, include_action_runs).get_repr()
            for job in self._master_control.jobs.itervalues()
        ]

    def render_GET(self, request):
        include_job_runs = requestargs.get_bool(request, 'include_job_runs')
        include_action_runs = requestargs.get_bool(request, 'include_action_runs')
        output = dict(jobs=self.get_data(include_job_runs, include_action_runs))
        return respond(request, output)

    def render_POST(self, request):
        cmd = requestargs.get_string(request, 'command')
        log.info("Handling '%s' request on all jobs", cmd)

        if cmd == 'disableall':
            self._master_control.disable_all()
            return respond(request, {'result': "All jobs are now disabled"})

        if cmd == 'enableall':
            self._master_control.enable_all()
            return respond(request, {'result': "All jobs are now enabled"})

        log.warning("Unknown request command %s for all jobs", cmd)
        return respond(request, None, code=http.NOT_IMPLEMENTED)


class ServiceInstanceResource(resource.Resource):

    isLeaf = True

    def __init__(self, service_instance, master_control):
        self._service_instance = service_instance
        self._master_control = master_control
        resource.Resource.__init__(self)

    def render_POST(self, request):
        cmd = requestargs.get_string(request, 'command')
        log.info("Handling '%s' request on service %s",
                 cmd, self._service_instance.id)

        if cmd == 'stop':
            self._service_instance.stop()
            return respond(request, {'result': "Service instance stopping"})

        if cmd == 'zap':
            self._service_instance.zap()
            return respond(request, {'result': "Service instance zapped"})

        if cmd == 'start':
            try:
                self._service_instance.start()
            except service.InvalidStateError:
                msg = ("Failed to start: Service is already %s" %
                       self._service_instance.state)
                return respond(request, {'result': msg})

            return respond(request, {'result': "Service instance starting"})

        log.warning("Unknown request command %s for service %s", cmd,
                    self._service_instance.id)
        return respond(request, None, code=http.NOT_IMPLEMENTED)


class ServiceResource(resource.Resource):
    """A resource that describes a particular service"""
    def __init__(self, service, master_control):
        self._service = service
        self._master_control = master_control
        resource.Resource.__init__(self)

    def getChild(self, name, request):
        if name == '':
            return self
        if name == '_events':
            return EventResource(self._service)

        for instance in self._service.instances:
            if str(instance.instance_number) == str(name):
                return ServiceInstanceResource(instance, self._master_control)
        else:
            return resource.NoResource("Cannot find service '%s'" % name)

    def get_instance_data(self, instance):
        return {
            'id':           instance.id,
            'node':         instance.node.hostname if instance.node else None,
            'state':        instance.state.name,
        }

    # TODO: create an adapter
    def render_GET(self, request):
        instance_output = [
            self.get_instance_data(instance)
            for instance in self._service.instances
        ]

        output = {
            'name':         self._service.name,
            'state':        self._service.state.name.upper(),
            'count':        self._service.count,
            'command':      self._service.command,
            'instances':    instance_output,
            'node_pool':    [n.hostname for n in self._service.node_pool.nodes]
        }
        return respond(request, output)

    def render_POST(self, request):
        cmd = requestargs.get_string(request, 'command')
        log.info("Handling '%s' request on service %s",
                 cmd, self._service.name)

        if cmd == 'stop':
            self._service.stop()
            return respond(request, {'result': "Service stopping"})

        if cmd == 'zap':
            self._service.zap()
            return respond(request, {'result': "Service zapped"})

        if cmd == 'start':
            try:
                self._service.start()
            except service.InvalidStateError:
                msg = ("Failed to start: Service is already %s" %
                       self._service.state)
                return respond(request, {'result': msg})

            return respond(request, {'result': "Service starting"})

        log.warning("Unknown request command %s for service %s",
                    cmd, self._service.name)
        return respond(request, None, code=http.NOT_IMPLEMENTED)


class ServicesResource(resource.Resource):
    """Resource for all our daemon's services"""

    def __init__(self, master_control):
        self._master_control = master_control
        resource.Resource.__init__(self)

    def getChild(self, name, request):
        if name == '':
            return self

        found = self._master_control.services.get(name)
        if found is None:
            return resource.NoResource("Cannot find service '%s'" % name)

        return ServiceResource(found, self._master_control)

    def get_data(self):
        service_list = []
        for current_service in self._master_control.services.itervalues():
            try:
                status = current_service.state.name.upper()
            except Exception, e:
                log.error("Unexpected service state: %s" % e)
                status = "BROKEN"
            try:
                count = current_service.count
            except Exception, e:
                log.error("Unexpected service count: %s" % e)
                count = -1

            service_desc = {
                'name': current_service.name,
                'count': count,
                'href': "/services/%s" % urllib.quote(current_service.name),
                'status': status,
            }
            service_list.append(service_desc)

        return service_list

    def render_GET(self, request):
        request.setHeader("content-type", "text/json")
        return respond(request, dict(services=self.get_data()))


class ConfigResource(resource.Resource):
    """Resource for configuration changes"""

    isLeaf = True

    def __init__(self, master_control):
        self._master_control = master_control
        resource.Resource.__init__(self)

    def render_GET(self, request):
        return respond(request,
                       {'config': self._master_control.config_lines()})

    def render_POST(self, request):
        log.info("Handling reconfig request")
        new_config = requestargs.get_string(request, 'config')
        self._master_control.rewrite_config(new_config)

        response = {'status': "Active"}
        try:
            self._master_control.live_reconfig()
        except Exception, e:
            log.exception("Failure doing live reconfig")
            response['error'] = str(e)

        return respond(request, response)


class StatusResource(resource.Resource):

    isLeaf = True

    def __init__(self, master_control):
        self._master_control = master_control
        resource.Resource.__init__(self)

    def render_GET(self, request):
        return respond(request, {'status': "I'm alive biatch"})


class EventResource(resource.Resource):

    isLeaf = True

    def __init__(self, recordable):
        resource.Resource.__init__(self)
        self._recordable = recordable

    def render_GET(self, request):
        response_data = []

        recorder = event.EventManager.get_instance().get(self._recordable)
        if not recorder:
            return respond(request, dict(data=[]))

        for evt in recorder.list():
            entity_desc = "UNKNOWN" if not evt.entity else str(evt.entity)
            response_data.append({
                'level':        evt.level,
                'name':         evt.name,
                'entity':       entity_desc,
                'time':         evt.time
            })
        return respond(request, dict(data=response_data))


class RootResource(resource.Resource):
    def __init__(self, master_control):
        self._master_control = master_control
        resource.Resource.__init__(self)

        # Setup children
        self.putChild('jobs',       JobsResource(master_control))
        self.putChild('services',   ServicesResource(master_control))
        self.putChild('config',     ConfigResource(master_control))
        self.putChild('status',     StatusResource(master_control))
        self.putChild('events',     EventResource(master_control))

    def getChild(self, name, request):
        if name == '':
            return self
        return resource.Resource.getChild(self, name, request)

    def render_GET(self, request):
        request.setHeader("content-type", "text/json")

        # We're going to load a big response with a bunch of stuff we know
        # about this tron instance

        jobs_resource = self.children["jobs"]
        services_resource = self.children["services"]

        response = {
            'jobs':             jobs_resource.get_data(),
            'jobs_href':        request.uri + request.childLink('jobs'),
            'services':         services_resource.get_data(),
            'services_href':    request.uri + request.childLink('services'),
            'config_href':      request.uri + request.childLink('config'),
            'status_href':      request.uri + request.childLink('status'),
        }
        return respond(request, response)


if __name__ == '__main__':
    from twisted.internet import reactor
    from testify.utils import turtle
    master_control = turtle.Turtle()
    master_control.jobs = {
        'test_job': turtle.Turtle(name="test_job",
                                  node=turtle.Turtle(hostname="batch0")),
    }
    reactor.listenTCP(8082, server.Site(RootResource(master_control)))
    reactor.run()



import logging
from tron.serialize import filehandler

from tron.utils import state, timeutils

log = logging.getLogger(__name__)

class ActionCommand(object):
    """An ActionCommand encapsulates a runnable task that is passed to a node
    for execution.

    A Node calls:
      started   (when the command starts)
      exited    (when the command exits)
      write_<channel> (when output is received)
      done      (when the command is finished)
    """

    class ActionState(state.NamedEventState):
        pass

    COMPLETE    = ActionState('complete')
    FAILSTART   = ActionState('failstart')
    EXITING     = ActionState('exiting', close=COMPLETE)
    RUNNING     = ActionState('running', exit=EXITING)
    PENDING     = ActionState('pending', start=RUNNING, exit=FAILSTART)

    STDOUT      = '.stdout'
    STDERR      = '.stderr'

    def __init__(self, id, command, serializer=None):
        self.id             = id
        self.command        = command
        self.machine        = state.StateMachine(
                                initial_state=self.PENDING, delegate=self)
        self.exit_status    = None
        self.start_time     = None
        self.end_time       = None
        self.stdout         = filehandler.NullFileHandle
        self.stderr         = filehandler.NullFileHandle
        if serializer:
            self.stdout         = serializer.open(self.STDOUT)
            self.stderr         = serializer.open(self.STDERR)

    @property
    def state(self):
        return self.machine.state

    @property
    def attach(self):
        return self.machine.attach

    def started(self):
        if not self.machine.check('start'):
            return False
        self.start_time = timeutils.current_timestamp()
        return self.machine.transition('start')

    def exited(self, exit_status):
        if not self.machine.check('exit'):
            return False
        self.end_time    = timeutils.current_timestamp()
        self.exit_status = exit_status
        return self.machine.transition('exit')

    def write_stderr(self, value):
        self.stderr.write(value)

    def write_stdout(self, value):
        self.stdout.write(value)

    def done(self):
        if not self.machine.check('close'):
            return False
        self.stdout.close()
        self.stderr.close()
        return self.machine.transition('close')

    def handle_errback(self, result):
        """Handle an unexpected error while being run.  This will likely be
        an interval error. Cleanup the state of this AcctionCommand and log
        something useful for debugging.
        """
        log.error("Unknown failure for ActionCommand run %s: %s\n%s",
                self.id, self.command, str(result))
        self.exited(result)
        self.done()

    def __repr__(self):
        return "ActionCommand %s %s: %s" % (self.id, self.command, self.state)


"""
Format and color output for tron commands.
"""
from operator import itemgetter
import os
import sys


class Color(object):

    enabled = None
    colors = {
        'gray':                 '\033[90m',
        'cyan':                 '\033[91m',
        'green':                '\033[92m',
        'yellow':               '\033[93m',
        'blue':                 '\033[94m',
        'red':                  '\033[91m',
        'white':                '\033[99m',
        # h is for highlighted
        'hgray':                '\033[100m',
        'hcyan':                '\033[101m',
        'hgreen':               '\033[102m',
        'hyellow':              '\033[103m',
        'hblue':                '\033[104m',
        'hred':                 '\033[106m',
        'end':                  '\033[0m',
    }

    @classmethod
    def set(cls, color_name, text):
        if not cls.enabled or not color_name:
            return unicode(text)
        return cls.colors[color_name.lower()] + unicode(text) + cls.colors['end']


class TableDisplay(object):
    """Base class for displaying columns of data.  This class takes a list
    of dict objects and formats it so that it displays properly in fixed width
    columns.  Overlap is truncated.

    This class provides many hooks for customizing the output, including:
        - sorting of rows
        - building composite values from more then one field
        - custom formatting of a columns values
        - adding additional data after each row
        - coloring of header, columns, or rows

    The default output is:

        Banner
        Header
        Row
        (optional post row)
        Row
        (optional post row)
        ...
        Footer
    """

    columns = None
    fields = None
    widths = None
    colors = None
    title = None

    header_color = 'hgray'

    @property
    def max_first_col_width(self):
        return max(self.num_cols - sum(self.widths[1:]), 10)

    def __init__(self, options=None):
        self.out = []
        self.options = options
        self.num_cols = self.console_width()

    def console_width(self):
        if not sys.stdout.isatty():
            return 80
        return int(os.popen('stty size', 'r').read().split()[1])

    def banner(self):
        if not self.title:
            return
        title = self.title.capitalize()
        self.out.append("\n%s:" % title)
        if not self.rows():
            self.out.append("No %s\n" % title)

    def header(self):
        row = [
            self.format_header(i, label)
            for i, label in enumerate(self.columns)
        ]
        self.out.append(
            Color.set(self.header_color, "".join(row))
        )

    def footer(self):
        pass

    def color(self, col, field):
        return None

    def sorted_fields(self, values):
        return [values[name] for name in self.fields]

    def format_row(self, fields):
        row = []
        for i, value in enumerate(self.sorted_fields(fields)):
            row.append(
                Color.set(self.color(i, value), self.format_value(i, value))
            )
        return Color.set(self.row_color(fields), "".join(row))

    def format_value(self, field, value):
        length = self.widths[field]
        value = unicode(value)
        if len(value) > length:
            return (value[:length - 3] + '...').ljust(length)
        return value.ljust(length)

    format_header = format_value

    def calculate_col_widths(self):
        padding = 3
        label_width = len(self.columns[0]) + 1
        if not self.data:
            max_name_width = 1
        else:
            max_name_width = max(len(item[self.fields[0]]) for item in self.data)
            max_name_width += padding
        self.widths[0] = min(
            self.max_first_col_width,
            max(label_width, max_name_width)
        )

    def output(self):
        out = "\n".join(self.out)
        self.out = []
        return out

    def post_row(self, row):
        pass

    def row_color(self, row):
        return None

    def rows(self):
        return sorted(self.data, key=itemgetter(self.fields[0]))

    def store_data(self, data):
        self.data = data

    def format(self, data):
        self.store_data(data)
        self.banner()

        self.calculate_col_widths()
        if not self.rows():
            return self.output()

        self.header()
        for row in self.rows():
            self.out.append(self.format_row(row))
            self.post_row(row)

        self.footer()
        return self.output()


class DisplayServices(TableDisplay):

    columns = ['Name',  'State',    'Count' ]
    fields  = ['name',  'status',   'count' ]
    widths  = [None,    10,         10      ]
    title = 'services'

    def format_details(self, service_content):
        self.out = [
            "Service: %s" % service_content['name'],
            "State: %s" % service_content['state'],
            "Instances: %r" % service_content['count'],
        ] + [
            "  %s : %s %s" % (serv['id'], serv['node'], serv['state'])
            for serv in service_content['instances']
        ]
        return self.output()

class DisplayJobRuns(TableDisplay):
    """Format Job runs."""

    columns = ['Run ID', 'State',    'Node', 'Scheduled Time']
    fields  = ['id',     'state',    'node', 'scheduled_time']
    widths  = [None,     6,          20,     25              ]
    title = 'job runs'

    def rows(self):
        data_rows = self.data
        if self.options.warn:
            warn_only_func = lambda r: r['state'] in ['FAIL', 'UNKWN', 'QUE']
            data_rows = filter(warn_only_func, self.data)
        return data_rows[:self.options.num_displays]

    def format_value(self, field, value):
        if self.fields[field] == 'id':
            value = '.' + '.'.join(value.split('.')[1:])

        return super(DisplayJobRuns, self).format_value(field, value)

    def sorted_fields(self, values):
        """Build constructed fields and return fields in order."""
        run = values['run_time'] or "-"
        values['scheduled_time'] = run

        return [values[name] for name in self.fields]

    def row_color(self, fields):
        return 'red' if fields['state'] == 'FAIL' else 'white'

    def post_row(self, row):
        start = row['start_time'] or "-"
        end =   row['end_time']   or "-"
        duration = row['duration'][:-7] if row['duration'] else "-"

        row_data = "%sStart: %s  End: %s  (%s)" % (
            ' ' * self.widths[0], start, end, duration
        )
        self.out.append(Color.set('gray', row_data))

        if self.options.warn:
            display_action = DisplayActions(self.options)
            self.out.append(display_action.format(row))


class DisplayJobs(TableDisplay):

    columns = ['Name',  'State',    'Scheduler',    'Last Success']
    fields  = ['name',  'status',   'scheduler',    'last_success']
    widths  = [None,    10,         20,             20            ]
    title = 'jobs'

    def post_row(self, row):
        if self.options.warn:
            self.out.extend(self.do_format_job(row, True))

    def format_job(self, job_details):
        self.out = self.do_format_job(job_details)
        return self.output()

    def do_format_job(self, job_details, supress_preface=False):
        out = []
        if self.options.display_preface and not supress_preface:
            out.extend([
                job_details['name'] + ":",
                "Scheduler: %s" % job_details['scheduler'],
                "\nList of Actions (topological):",
            ] + job_details['action_names'] + [
                "\nNode Pool:"
            ] + job_details['node_pool'] + [
                "\nRun History: (%d total)" % len(job_details['runs'])
            ])
        job_runs = self.format_job_runs(job_details['runs'])
        out.append(job_runs)
        return out

    def format_job_runs(self, runs):
        return DisplayJobRuns(self.options).format(runs)


class DisplayActions(TableDisplay):

    columns = ['Action', 'State', 'Start Time', 'End Time', 'Duration']
    fields  = ['id',     'state', 'start_time', 'end_time', 'duration']
    widths  = [None,     7,        22,          22,         10        ]
    title = 'actions'

    def banner(self):
        if self.options.display_preface:
            self.out.extend([
                "Job Run: %s" % self.action['id'],
                "State: %s" % self.action['state'],
                "Node: %s" % self.action['node'],
            ])
        super(DisplayActions, self).banner()

    def footer(self):
        if len(self.rows()) < len(self.data) and not self.options.warn:
            self.out.append('...')

    def format_value(self, field, value):
        if self.fields[field] == 'id':
            value = '.' + '.'.join(value.split('.')[2:])
        if self.fields[field] in ('start_time', 'end_time'):
            value = value or "-"
        if self.fields[field] == 'duration':
            # Strip microseconds
            value = value[:-7] if value else "-"

        return super(DisplayActions, self).format_value(field, value)

    def row_color(self, fields):
        return 'red' if fields['state'] == 'FAIL' else 'white'

    def store_data(self, data):
        self.data = data['runs']
        self.action = data

    def rows(self):
        data_rows = self.data
        if self.options.warn:
            warn_only_func = lambda r: r['state'] in ['FAIL', 'UNKWN', 'QUE']
            data_rows = filter(warn_only_func, self.data)
        return data_rows[:self.options.num_displays]

    def post_row(self, row):
        if self.options.warn:
            self.out.extend(self.do_format_action_run(row, True))

    def format_action_run(self, content):
        self.out = self.do_format_action_run(content)
        return self.output()

    def do_format_action_run(self, content, supress_preface=False):
        out = []
        if self.options.stdout:
            out.extend(["Stdout: "] + content['stdout'])
            return out

        if self.options.stderr or self.options.warn:
            out.extend(["Stderr: "] + content['stderr'])
            return out

        if self.options.display_preface and not supress_preface:
            out.extend([
                "Action Run: %s" % content['id'],
                "State: %s" % content['state'],
                "Node: %s" % content['node'],
                ''
            ])

        # a raw command is without command context
        if content['command'] != content['raw_command']:
            if content['command'] == 'false':
                out.append(Color.set("red", "Bad Command"))
            else:
                out.append(Color.set("gray", content['command']))

        out.extend(
            [Color.set("gray", content['raw_command'])] +
            ["\nRequirements:"] + content['requirements'] +
            ["\nStdout:"] + content['stdout'] +
            ["\nStderr:"] + content['stderr']
        )
        return out


class DisplayEvents(TableDisplay):

    columns = ['Time', 'Level', 'Entity', 'Name']
    fields  = ['time', 'level', 'entity', 'name']
    widths  = [22,     12,       35,      20    ]
    title = 'events'

    def calculate_col_widths(self):
        # No need to calculate, it's fixed width.
        pass


def view_with_less(content, color=True):
    """Send `content` through less."""
    import subprocess
    cmd = ['less']
    if color:
        cmd.append('-r')

    less_proc = subprocess.Popen(cmd, stdin=subprocess.PIPE)
    less_proc.stdin.write(content)
    less_proc.stdin.close()
    less_proc.wait()


import os
import pwd
import struct
import logging

from twisted.internet import defer
from twisted.conch.ssh import channel, common
from twisted.conch.ssh import connection
from twisted.conch.ssh import transport
from twisted.conch.client import default
from twisted.python import failure

log = logging.getLogger('tron.ssh')


class Error(Exception):
    pass


class ChannelClosedEarlyError(Error):
    """Indicates the SSH Channel has closed before we were done handling the
    command"""
    pass


# We need to sub-class and redefine the AuthClient here because there is no way
# with the default client to force it to not try certain authentication
# methods. If things get much worse I'll just have to make our own custom
# version of the default module.
class NoPasswordAuthClient(default.SSHUserAuthClient):
    def tryAuth(self, kind):
        kind = kind.replace('-', '_')
        if kind != 'publickey':
            log.info('skipping auth method %s (not supported)' % kind)
            return

        log.info('trying to auth with %s!' % kind)
        f = getattr(self, 'auth_%s' % kind, None)
        if f:
            return f()
        else:
            return


class ClientTransport(transport.SSHClientTransport):

    connection_defer = None

    def __init__(self, *args, **kwargs):
        # These silly twisted classes tend not to have init functions, and
        # they're all old style classes
        # transport.SSHClientTransport.__init__(self, *args, **kwargs)

        self.options = kwargs['options']

    def verifyHostKey(self, pubKey, fingerprint):
        return defer.succeed(1)

    def connectionSecure(self):
        conn = ClientConnection()
        conn.service_defer = defer.Deferred()

        self.connection_defer.callback(conn)

        auth_service = NoPasswordAuthClient(pwd.getpwuid(os.getuid())[0],
                                            self.options, conn)

        self.requestService(auth_service)


class ClientConnection(connection.SSHConnection):

    service_start_defer = None
    service_stop_defer = None

    def serviceStarted(self):
        log.info("Service started")
        connection.SSHConnection.serviceStarted(self)
        if not self.service_stop_defer.called:
            self.service_start_defer.callback(self)

    def serviceStopped(self):
        log.info("Service stopped")
        connection.SSHConnection.serviceStopped(self)
        if not self.service_stop_defer.called:
            self.service_stop_defer.callback(self)

    def channelClosed(self, channel):
        if not channel.conn:
            log.warning("Channel %r failed to open", channel.id)
            # Channel has no connection, so we were still trying to open it The
            # normal error handling won't notify us since the channel never
            # successfully opened.
            channel.openFailed(None)

        connection.SSHConnection.channelClosed(self, channel)


class ExecChannel(channel.SSHChannel):

    name = 'session'
    exit_defer = None
    start_defer = None

    command = None
    exit_status = None
    running = False

    def __init__(self, *args, **kwargs):
        channel.SSHChannel.__init__(self, *args, **kwargs)
        self.output_callbacks = []
        self.end_callbacks = []
        self.error_callbacks = []
        self.data = []

    def channelOpen(self, data):
        self.data = []
        self.running = True

        if self.start_defer:
            log.debug("Channel %s is open, calling deferred", self.id)
            self.start_defer.callback(self)
            req = self.conn.sendRequest(self, 'exec',
                                        common.NS(self.command),
                                        wantReply=True)
            req.addCallback(self._cbExecSendRequest)
        else:
            # A missing start defer means that we are no longer expected to do
            # anything when the channel opens It probably means we gave up on
            # this connection and failed the job, but later the channel opened
            # up correctly.
            log.warning("Channel open delayed, giving up and closing")
            self.loseConnection()

    def addOutputCallback(self, output_callback):
        self.output_callbacks.append(output_callback)

    def addErrorCallback(self, error_callback):
        self.error_callbacks.append(error_callback)

    def addEndCallback(self, end_callback):
        self.end_callbacks.append(end_callback)

    def openFailed(self, reason):
        log.error("Open failed due to %r", reason)
        if self.start_defer:
            self.start_defer.errback(self)

    def _cbExecSendRequest(self, ignored):
        self.conn.sendEOF(self)

    def request_exit_status(self, data):
        # exit status is a 32-bit unsigned int in network byte format
        status = struct.unpack_from(">L", data, 0)[0]

        log.debug("Received exit status request: %d", status)
        self.exit_status = status
        self.exit_defer.callback(self)
        self.running = False
        return True

    def dataReceived(self, data):
        self.data.append(data)
        for callback in self.output_callbacks:
            callback(data)

    def extReceived(self, dataType, data):
        self.data.append(data)
        for callback in self.error_callbacks:
            callback(data)

    def getStdout(self):
        return "".join(self.data)

    def closed(self):
        if (self.exit_status is None and
            self.running and
            self.exit_defer and
            not self.exit_defer.called):
            log.warning("Channel has been closed without receiving an exit"
                        " status")
            f = failure.Failure(exc_value=ChannelClosedEarlyError())
            self.exit_defer.errback(f)

        for callback in self.end_callbacks:
            callback()
        self.loseConnection()



#!/usr/bin/env python
#
# Copyright 2007 Google Inc.
# Modified 2011 Yelp Inc.
# from http://code.google.com/p/googleappengine/source/browse/trunk/python/google/appengine/cron/groctimespecification.py
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Implementation of scheduling for Groc format schedules.

A Groc schedule looks like '1st,2nd monday 9:00', or 'every 20 mins'. This
module takes a parsed schedule (produced by Antlr) and creates objects that
can produce times that match this schedule.

A parsed schedule is one of two types - an Interval or a Specific Time.
See the class docstrings for more.

Extensions to be considered:

  allowing a comma separated list of times to run
"""


import calendar
import datetime

try:
  import pytz
  assert pytz
except ImportError:
  pytz = None

HOURS = 'hours'
MINUTES = 'minutes'

try:
  from pytz import NonExistentTimeError
  assert NonExistentTimeError
  from pytz import AmbiguousTimeError
  assert AmbiguousTimeError
except ImportError:
  class NonExistentTimeError(Exception):
    pass

  class AmbiguousTimeError(Exception):
    pass

class GrocException(Exception):
    pass


class TimeSpecification(object):
  """Base class for time specifications."""

  def GetMatches(self, start, n):
    """Returns the next n times that match the schedule, starting at time start.

    Arguments:
      start: a datetime to start from. Matches will start from after this time.
      n:     the number of matching times to return

    Returns:
      a list of n datetime objects
    """
    out = []
    for _ in range(n):
      start = self.GetMatch(start)
      out.append(start)
    return out

  def GetMatch(self, start):
    """Returns the next match after time start.

    Must be implemented in subclasses.

    Arguments:
      start: a datetime to start from. Matches will start from after this time.
          This may be in any pytz time zone, or it may be timezone-naive
          (interpreted as UTC).

    Returns:
      a datetime object in the timezone of the input 'start'
    """
    raise NotImplementedError


def _GetTimezone(timezone_string):
  """Converts a timezone string to a pytz timezone object.

  Arguments:
    timezone_string: a string representing a timezone, or None

  Returns:
    a pytz timezone object, or None if the input timezone_string is None

  Raises:
    ValueError: if timezone_string is not None and the pytz module could not be
        loaded
  """
  if timezone_string:
    if pytz is None:
      raise ValueError('need pytz in order to specify a timezone')
    return pytz.timezone(timezone_string)
  else:
    return None


def _ToTimeZone(t, tzinfo):
  """Converts 't' to the time zone 'tzinfo'.

  Arguments:
    t: a datetime object.  It may be in any pytz time zone, or it may be
        timezone-naive (interpreted as UTC).
    tzinfo: a pytz timezone object, or None (interpreted as UTC).

  Returns:
    a datetime object in the time zone 'tzinfo'
  """
  if pytz is None:
    return t.replace(tzinfo=tzinfo)
  elif tzinfo:
    if not t.tzinfo:
      t = pytz.utc.localize(t)
    return tzinfo.normalize(t.astimezone(tzinfo))
  elif t.tzinfo:
    return pytz.utc.normalize(t.astimezone(pytz.utc)).replace(tzinfo=None)
  else:
    return t


def _GetTime(time_string):
  """Converts a string to a datetime.time object.

  Arguments:
    time_string: a string representing a time ('hours:minutes')

  Returns:
    a datetime.time object
  """
  try:
    hourstr, minutestr = time_string.split(':')[0:2]
    return datetime.time(int(hourstr), int(minutestr))
  except ValueError:
    return None

class IntervalTimeSpecification(TimeSpecification):
  """A time specification for a given interval.

  An Interval type spec runs at the given fixed interval. It has the following
  attributes:
  period - the type of interval, either 'hours' or 'minutes'
  interval - the number of units of type period.
  synchronized - whether to synchronize the times to be locked to a fixed
      period (midnight in the specified timezone).
  start_time, end_time - restrict matches to a given range of times every day.
      If these are None, there is no restriction.  Otherwise, they are
      datetime.time objects.
  timezone - the time zone in which start_time and end_time should be
      interpreted, or None (defaults to UTC).  This is a pytz timezone object.
  """
  def __init__(self, interval, period, synchronized=False,
               start_time_string='', end_time_string='', timezone=None):
    super(IntervalTimeSpecification, self).__init__()
    if interval < 1:
      raise GrocException('interval must be greater than zero')
    self.interval = interval
    self.period = period
    self.synchronized = synchronized
    if self.period == HOURS:
      self.seconds = self.interval * 3600
    else:
      self.seconds = self.interval * 60
    self.timezone = _GetTimezone(timezone)

    if self.synchronized:
      if start_time_string:
        raise ValueError(
            'start_time_string may not be specified if synchronized is true')
      if end_time_string:
        raise ValueError(
            'end_time_string may not be specified if synchronized is true')
      if (self.seconds > 86400) or ((86400 % self.seconds) != 0):
        raise GrocException('can only use synchronized for periods that'
                                 ' divide evenly into 24 hours')

      self.start_time = datetime.time(0, 0).replace(tzinfo=self.timezone)
      self.end_time = datetime.time(23, 59).replace(tzinfo=self.timezone)
    elif start_time_string:
      if not end_time_string:
        raise ValueError(
            'end_time_string must be specified if start_time_string is')
      self.start_time = (
          _GetTime(start_time_string).replace(tzinfo=self.timezone))
      self.end_time = _GetTime(end_time_string).replace(tzinfo=self.timezone)
    else:
      if end_time_string:
        raise ValueError(
            'start_time_string must be specified if end_time_string is')
      self.start_time = None
      self.end_time = None

  def GetMatch(self, start):
    """Returns the next match after 'start'.

    Arguments:
      start: a datetime to start from. Matches will start from after this time.
          This may be in any pytz time zone, or it may be timezone-naive
          (interpreted as UTC).

    Returns:
      a datetime object in the timezone of the input 'start'
    """
    if self.start_time is None:
      return start + datetime.timedelta(seconds=self.seconds)

    t = _ToTimeZone(start, self.timezone)

    start_time = self._GetPreviousDateTime(t, self.start_time)

    t_delta = t - start_time
    t_delta_seconds = (t_delta.days * 60 * 24 + t_delta.seconds)
    num_intervals = (t_delta_seconds + self.seconds) / self.seconds
    interval_time = (
        start_time + datetime.timedelta(seconds=(num_intervals * self.seconds)))
    if self.timezone:
      interval_time = self.timezone.normalize(interval_time)

    next_start_time = self._GetNextDateTime(t, self.start_time)
    if (self._TimeIsInRange(t) and
        self._TimeIsInRange(interval_time) and
        interval_time < next_start_time):
      result = interval_time
    else:
      result = next_start_time


    return _ToTimeZone(result, start.tzinfo)

  def _TimeIsInRange(self, t):
    """Returns true if 't' falls between start_time and end_time, inclusive.

    Arguments:
      t: a datetime object, in self.timezone

    Returns:
      a boolean
    """
    previous_start_time = self._GetPreviousDateTime(t, self.start_time)
    previous_end_time = self._GetPreviousDateTime(t, self.end_time)
    if previous_start_time > previous_end_time:
      return True
    else:
      return t == previous_end_time

  @staticmethod
  def _GetPreviousDateTime(t, target_time):
    """Returns the latest datetime <= 't' that has the time target_time.

    Arguments:
      t: a datetime.datetime object, in self.timezone
      target_time: a datetime.time object, in self.timezone

    Returns:
      a datetime.datetime object, in self.timezone
    """
    date = t.date()
    while True:
      result = IntervalTimeSpecification._CombineDateAndTime(date, target_time)
      if result <= t:
        return result
      date -= datetime.timedelta(days=1)

  @staticmethod
  def _GetNextDateTime(t, target_time):
    """Returns the earliest datetime > 't' that has the time target_time.

    Arguments:
      t: a datetime.datetime object, in self.timezone
      target_time: a time object, in self.timezone

    Returns:
      a datetime.datetime object, in self.timezone
    """
    date = t.date()
    while True:
      result = IntervalTimeSpecification._CombineDateAndTime(date, target_time)
      if result > t:
        return result
      date += datetime.timedelta(days=1)

  @staticmethod
  def _CombineDateAndTime(date, time):
    """Creates a datetime object from date and time objects.

    This is similar to the datetime.combine method, but its timezone
    calculations are designed to work with pytz.

    Arguments:
      date: a datetime.date object, in any timezone
      time: a datetime.time object, in any timezone

    Returns:
      a datetime.datetime object, in the timezone of the input 'time'
    """
    if time.tzinfo:
      naive_result = datetime.datetime(
          date.year, date.month, date.day, time.hour, time.minute, time.second)
      try:
        return time.tzinfo.localize(naive_result, is_dst=None)
      except AmbiguousTimeError:


        return min(time.tzinfo.localize(naive_result, is_dst=True),
                   time.tzinfo.localize(naive_result, is_dst=False))
      except NonExistentTimeError:
        while True:
          naive_result += datetime.timedelta(minutes=1)
          try:
            return time.tzinfo.localize(naive_result, is_dst=None)
          except NonExistentTimeError:
            pass
    else:
      return datetime.datetime.combine(date, time)


class SpecificTimeSpecification(TimeSpecification):
  """Specific time specification.

  A Specific interval is more complex, but defines a certain time to run and
  the days that it should run. It has the following attributes:
  time     - the time of day to run, as 'HH:MM'
  ordinals - first, second, third &c, as a set of integers in 1..5
  months   - the months that this should run, as a set of integers in 1..12
  weekdays - the days of the week that this should run, as a set of integers,
             0=Sunday, 6=Saturday
  timezone - the optional timezone as a string for this specification.
             Defaults to UTC - valid entries are things like Australia/Victoria
             or PST8PDT.

  A specific time schedule can be quite complex. A schedule could look like
  this:
  '1st,third sat,sun of jan,feb,mar 09:15'

  In this case, ordinals would be {1,3}, weekdays {0,6}, months {1,2,3} and
  time would be '09:15'.
  """

  def __init__(self, ordinals=None, weekdays=None, months=None, monthdays=None,
               timestr='00:00', timezone=None):

    super(SpecificTimeSpecification, self).__init__()
    if weekdays and monthdays:
      raise ValueError('cannot supply both monthdays and weekdays')
    if ordinals is None:

      self.ordinals = set(range(1, 6))
    else:
      self.ordinals = set(ordinals)

    if weekdays is None:

      self.weekdays = set(range(7))
    else:
      self.weekdays = set(weekdays)

    if months is None:

      self.months = set(range(1, 13))
    else:
      self.months = set(months)

    if not monthdays:
      self.monthdays = set()
    else:
      if max(monthdays) > 31 or min(monthdays) < 1:
        raise ValueError('invalid day of month')
      self.monthdays = set(monthdays)
    self.time = _GetTime(timestr)
    self.timezone = _GetTimezone(timezone)

  def _MatchingDays(self, year, month):
    """Returns matching days for the given year and month.

    For the given year and month, return the days that match this instance's
    day specification, based on either (a) the ordinals and weekdays, or
    (b) the explicitly specified monthdays.  If monthdays are specified,
    dates that fall outside the range of the month will not be returned.

    Arguments:
      year: the year as an integer
      month: the month as an integer, in range 1-12

    Returns:
      a list of matching days, as ints in range 1-31
    """
    start_day, last_day = calendar.monthrange(year, month)
    if self.monthdays:
      return sorted([day for day in self.monthdays if day <= last_day])

    out_days = []
    start_day = (start_day + 1) % 7
    for ordinal in self.ordinals:
      for weekday in self.weekdays:
        day = ((weekday - start_day) % 7) + 1
        day += 7 * (ordinal - 1)
        if day <= last_day:
          out_days.append(day)
    return sorted(out_days)

  def _NextMonthGenerator(self, start, matches):
    """Creates a generator that produces results from the set 'matches'.

    Matches must be >= 'start'. If none match, the wrap counter is incremented,
    and the result set is reset to the full set. Yields a 2-tuple of (match,
    wrapcount).

    Arguments:
      start: first set of matches will be >= this value (an int)
      matches: the set of potential matches (a sequence of ints)

    Yields:
      a two-tuple of (match, wrap counter). match is an int in range (1-12),
      wrapcount is a int indicating how many times we've wrapped around.
    """
    potential = matches = sorted(matches)

    after = start - 1
    wrapcount = 0
    while True:
      potential = [x for x in potential if x > after]
      if not potential:

        wrapcount += 1
        potential = matches
      after = potential[0]
      yield (after, wrapcount)

  def GetMatch(self, start):
    """Returns the next match after time start.

    Must be implemented in subclasses.

    Arguments:
      start: a datetime to start from. Matches will start from after this time.
          This may be in any pytz time zone, or it may be timezone-naive
          (interpreted as UTC).

    Returns:
      a datetime object in the timezone of the input 'start'
    """
    start_time = _ToTimeZone(start, self.timezone).replace(tzinfo=None)
    if self.months:

      months = self._NextMonthGenerator(start_time.month, self.months)
    while True:

      month, yearwraps = months.next()
      candidate_month = start_time.replace(day=1, month=month,
                                           year=start_time.year + yearwraps)

      day_matches = self._MatchingDays(candidate_month.year, month)

      if ((candidate_month.year, candidate_month.month)
          == (start_time.year, start_time.month)):

        day_matches = [x for x in day_matches if x >= start_time.day]

        while (day_matches and day_matches[0] == start_time.day
               and start_time.time() >= self.time):
          day_matches.pop(0)
      while day_matches:

        out = candidate_month.replace(day=day_matches[0], hour=self.time.hour,
                                      minute=self.time.minute, second=0,
                                      microsecond=0)

        if self.timezone and pytz is not None:
          try:
            out = self.timezone.localize(out, is_dst=None)
          except AmbiguousTimeError:
            out = self.timezone.localize(out)
          except NonExistentTimeError:
            for _ in range(24):
              out = out + datetime.timedelta(minutes=60)
              try:
                out = self.timezone.localize(out)
              except NonExistentTimeError:
                continue
              break
        return _ToTimeZone(out, start.tzinfo)




"""
A command line http client used by tronview, tronctl, and tronfig
"""
import logging
import urllib
import urllib2
import urlparse

try:
    import simplejson
    assert simplejson # Pyflakes
except ImportError:
    import json as simplejson

log = logging.getLogger("tron.commands.client")

USER_AGENT = "Tron Command/1.0 +http://github.com/Yelp/Tron"

# Result Codes
OK = "OK"
REDIRECT = "REDIRECT"
ERROR = "ERROR"


def request(host, path, data=None):
    enc_data = urllib.urlencode(data) if data else data

    uri = urlparse.urljoin(host, path)
    req = urllib2.Request(uri, enc_data)
    log.info("Request to %r", uri)

    req.add_header("User-Agent", USER_AGENT)
    opener = urllib2.build_opener()
    try:
        output = opener.open(req)
    except urllib2.HTTPError, e:
        log.error("Recieved error response: %s" % e)
        return ERROR, e.code
    except urllib2.URLError, e:
        log.error("Recieved error response: %s" % e)
        return ERROR, e.reason

    result = simplejson.load(output)
    return OK, result


class Client(object):
    """A client used in commands to make requests to the tron.www """

    def __init__(self, options):
        self.options = options

    def status(self):
        return self.request('/status')

    def events(self):
        return self.request('/events')['data']

    def config(self, data=None):
        """This may be a post or a get, depending on data."""
        if data:
            return self.request('/config', dict(config=data))
        return self.request('/config', )['config']

    def index(self):
        content = self.request('/')

        def name_href_dict(source):
            return dict((i['name'], i['href']) for i in source)

        return {
            'jobs':     name_href_dict(content['jobs']),
            'services': name_href_dict(content['services'])
        }

    def get_url_from_identifier(self, iden):
        """Convert a string of the form job_name[.run_number[.action]] to its
        corresponding URL.
        """
        obj_name_elements = iden.split('.')
        obj_name = obj_name_elements[0]
        obj_rel_path = '/'.join(obj_name_elements[1:])

        def full_url(obj_url):
            return '/'.join((obj_url, obj_rel_path))

        content = self.index()
        if obj_name in content['jobs']:
            return full_url(content['jobs'][obj_name])
        if obj_name in content['services']:
            return full_url(content['services'][obj_name])

        raise ValueError("Unknown identifier: %s" % iden)

    def services(self):
        return self.request('/services').get('services')

    def service(self, service_id):
        service_url = "/services/%s" % service_id
        return self.request(service_url)

    def service_events(self, service_id):
        service_url = "/services/%s/_events" % service_id
        return self.request(service_url)['data']

    def _get_job_params(self):
        if self.options.warn:
            return "?include_job_runs=1&include_action_runs=1"
        return ''

    def jobs(self):
        params = self._get_job_params()
        return self.request('/jobs' + params).get('jobs')

    def job(self, job_id):
        params = self._get_job_params()
        return self.request('/jobs/%s%s' % (job_id, params))

    def job_events(self, job_id):
        return self.request('/jobs/%s/_events' % job_id)['data']

    def job_runs(self, action_id):
        params = self._get_job_params()
        action_id = action_id.replace('.', '/')
        return self.request('/jobs/%s%s' % (action_id, params))

    def action(self, action_id):
        url = "/jobs/%s?num_lines=%s" % (action_id.replace('.', '/'),
            self.options.num_displays)
        return self.request(url)

    def action_events(self, action_id):
        action_id = action_id.replace('.', '/')
        return self.request('/jobs/%s/_events' % action_id)['data']

    def request(self, url, data=None):
        status, content = request(self.options.server, url, data)
        if not status == OK:
            err_msg = "Failed to request %s%s: %s %s" % (
                self.options.server, url, content, data or '')
            raise ValueError(err_msg)
        return content



import logging
from twisted.internet import reactor

from tron import command_context, event, node
from tron.core import jobrun
from tron.core import actiongraph
from tron.core.actionrun import ActionRun
from tron.serialize import filehandler
from tron.utils import timeutils
from tron.utils.observer import Observable, Observer

class Error(Exception):
    pass


class ConfigBuildMismatchError(Error):
    pass


class InvalidStartStateError(Error):
    pass


log = logging.getLogger(__name__)


class JobContext(object):
    """A class which exposes properties for rendering commands."""

    def __init__(self, job):
        self.job = job

    @property
    def name(self):
        return self.job.name

    def __getitem__(self, item):
        date_name, date_spec = self._get_date_spec_parts(item)
        if not date_spec:
            raise KeyError(item)

        if date_name == 'last_success':
            last_success = self.job.runs.last_success
            last_success = last_success.run_time if last_success else None

            time_value = timeutils.DateArithmetic.parse(date_spec, last_success)
            if time_value:
                return time_value

        raise KeyError(item)

    def _get_date_spec_parts(self, name):
        parts = name.rsplit(':', 1)
        if len(parts) != 2:
            return [name, None]
        return parts


class Job(Observable, Observer):
    """A configurable data object.

    Job uses JobRunCollection to manage its runs, and ActionGraph to manage its
    actions and their dependency graph.
    """

    STATUS_DISABLED       = "DISABLED"
    STATUS_ENABLED        = "ENABLED"
    STATUS_UNKNOWN        = "UNKNOWN"
    STATUS_RUNNING        = "RUNNING"

    NOTIFY_STATE_CHANGE   = 'notify_state_change'
    NOTIFY_RUN_DONE       = 'notify_run_done'

    EVENT_RECONFIGURED    = event.EventType(event.LEVEL_NOTICE, 'reconfigured')
    EVENT_RUN_CREATED     = event.EventType(event.LEVEL_NOTICE, 'run_created')
    EVENT_STATE_RESTORED  = event.EventType(event.LEVEL_INFO, 'restored')

    def __init__(self, name, scheduler, queueing=True, all_nodes=False,
            node_pool=None, enabled=True, action_graph=None,
            run_collection=None, parent_context=None, output_path=None):
        super(Job, self).__init__()
        self.name               = name
        self.action_graph       = action_graph
        self.scheduler          = scheduler
        self.runs               = run_collection
        self.queueing           = queueing
        self.all_nodes          = all_nodes
        self.enabled            = enabled
        self.node_pool          = node_pool
        self.context            = command_context.CommandContext(
                                    JobContext(self), parent_context)
        self.output_path        = output_path or filehandler.OutputPath()
        self.output_path.append(name)

    @classmethod
    def from_config(cls, job_config, scheduler, parent_context, output_path):
        """Factory method to create a new Job instance from configuration."""
        node_pools = node.NodePoolStore.get_instance()
        action_graph = actiongraph.ActionGraph.from_config(
                job_config.actions, node_pools, job_config.cleanup_action)
        runs = jobrun.JobRunCollection.from_config(job_config)
        nodes = node_pools[job_config.node] if job_config.node else None

        return cls(
            name                = job_config.name,
            queueing            = job_config.queueing,
            all_nodes           = job_config.all_nodes,
            node_pool           = nodes,
            scheduler           = scheduler,
            enabled             = job_config.enabled,
            run_collection      = runs,
            action_graph        = action_graph,
            parent_context      = parent_context,
            output_path         = output_path
        )

    def update_from_job(self, job):
        """Update this Jobs configuration from a new config. This method
        actually takes an already constructed job and copies out its
        configuration data.
        """
        self.name           = job.name
        self.queueing       = job.queueing
        self.scheduler      = job.scheduler
        self.node_pool      = job.node_pool
        self.all_nodes      = job.all_nodes
        self.action_graph   = job.action_graph
        self.enabled        = job.enabled
        self.output_path    = job.output_path
        self.context        = job.context
        self.notify(self.EVENT_RECONFIGURED)

    @property
    def status(self):
        """Current status."""
        if not self.enabled:
            return self.STATUS_DISABLED
        if self.runs.get_run_by_state(ActionRun.STATE_RUNNING):
            return self.STATUS_RUNNING

        if (self.runs.get_run_by_state(ActionRun.STATE_SCHEDULED) or
                self.runs.get_run_by_state(ActionRun.STATE_QUEUED)):
            return self.STATUS_ENABLED

        log.warn("%s in an unknown state: %s" % (self, self.runs))
        return self.STATUS_UNKNOWN

    @property
    def state_data(self):
        """This data is used to serialize the state of this job."""
        return {
            'runs':             self.runs.state_data,
            'enabled':          self.enabled
        }

    def restore_state(self, state_data):
        """Apply a previous state to this Job."""
        self.enabled = state_data['enabled']
        job_runs = self.runs.restore_state(
                state_data['runs'],
                self.action_graph,
                self.output_path.clone(),
                self.context,
                self.node_pool)
        for run in job_runs:
            self.watch(run)

        self.notify(self.EVENT_STATE_RESTORED)

    def build_new_runs(self, run_time, manual=False):
        """Uses its JobCollection to build new JobRuns. If all_nodes is set,
        build a run for every node, otherwise just builds a single run on a
        single node.
        """
        pool = self.node_pool
        nodes = pool.nodes if self.all_nodes else [pool.next()]
        for node in nodes:
            run = self.runs.build_new_run(self, run_time, node, manual=manual)
            self.watch(run)
            event.EventManager.get_instance().add(run, parent=self)
            self.notify(self.EVENT_RUN_CREATED)
            yield run

    def handle_job_run_state_change(self, job_run, event):
        """Handle state changes from JobRuns and propagate changes to any
        observers.
        """
        # Propagate state change for serialization
        if event == jobrun.JobRun.NOTIFY_STATE_CHANGED:
            self.notify(self.NOTIFY_STATE_CHANGE)
            return

        # Propagate DONE JobRun notifications to JobScheduler
        if event == jobrun.JobRun.NOTIFY_DONE:
            self.notify(self.NOTIFY_RUN_DONE)
            return
    handler = handle_job_run_state_change

    def __eq__(self, other):
        attrs = [
                'name',
                'queueing',
                'scheduler',
                'node_pool',
                'all_nodes',
                'action_graph',
                'output_path',
        ]
        return all(
            getattr(other, attr, None) == getattr(self, attr, None)
            for attr in attrs
        )

    def __ne__(self, other):
        return not self == other

    def __str__(self):
        return "Job:%s" % self.name


class JobScheduler(Observer):
    """A JobScheduler is responsible for scheduling Jobs and running JobRuns
    based on a Jobs configuration. Runs jobs by setting a callback to fire
    x seconds into the future.
    """

    def __init__(self, job):
        self.job = job
        self.watch(job)

    def restore_job_state(self, job_state_data):
        """Restore the job state and schedule any JobRuns."""
        self.job.restore_state(job_state_data)
        scheduled = self.job.runs.get_scheduled()
        for job_run in scheduled:
            self._set_callback(job_run)
        # Ensure we have at least 1 scheduled run
        self.schedule()

    def enable(self):
        """Enable the job and start its scheduling cycle."""
        if self.job.enabled:
            return

        self.job.enabled = True
        for job_run in self.get_runs_to_schedule(ignore_last_run_time=True):
            self._set_callback(job_run)

    def disable(self):
        """Disable the job and cancel and pending scheduled jobs."""
        self.job.enabled = False
        self.job.runs.cancel_pending()

    def manual_start(self, run_time=None):
        """Trigger a job run manually (instead of from the scheduler)."""
        run_time = run_time or timeutils.current_time()
        manual_runs = list(self.job.build_new_runs(run_time, manual=True))
        for r in manual_runs:
            r.start()
        return manual_runs

    def schedule_reconfigured(self):
        """Remove the pending run and create new runs with the new JobScheduler.
        """
        self.job.runs.remove_pending()
        self.schedule()

    def schedule(self):
        """Schedule the next run for this job by setting a callback to fire
        at the appropriate time.
        """
        if not self.job.enabled:
            return

        for job_run in self.get_runs_to_schedule():
            self._set_callback(job_run)

    def _set_callback(self, job_run):
        """Set a callback for JobRun to fire at the appropriate time."""
        log.info("Scheduling next Jobrun for %s", self.job.name)
        seconds = job_run.seconds_until_run_time()
        reactor.callLater(seconds, self.run_job, job_run)

    def run_job(self, job_run, run_queued=False):
        """Triggered by a callback to actually start the JobRun. Also
        schedules the next JobRun.
        """
        # If the Job has been disabled after this run was scheduled, then cancel
        # the JobRun and do not schedule another
        if not self.job.enabled:
            log.info("%s cancelled because job has been disabled." % job_run)
            return job_run.cancel()

        # If the JobRun was cancelled we won't run it.  A JobRun may be
        # cancelled if the job was disabled, or manually by a user. It's
        # also possible this job was run (or is running) manually by a user.
        # Alternatively, if run_queued is True, this job_run is already queued.
        if not run_queued and not job_run.is_scheduled:
            log.info("%s in state %s already out of scheduled state." % (
                    job_run, job_run.state))
            return self.schedule()

        node = job_run.node if self.job.all_nodes else None
        # If there is another job run still running, queue or cancel this one
        if any(self.job.runs.get_active(node)):
            self._queue_or_cancel_active(job_run)
            return

        job_run.start()
        if not self.job.scheduler.schedule_on_complete:
            self.schedule()

    def _queue_or_cancel_active(self, job_run):
        if self.job.queueing:
            log.info("%s still running, queueing %s." % (self.job, job_run))
            return job_run.queue()

        log.info("%s still running, cancelling %s." % (self.job, job_run))
        job_run.cancel()
        self.schedule()

    def handle_job_events(self, _observable, event):
        """Handle notifications from observables. If a JobRun has completed
        look for queued JobRuns that may need to start now.
        """
        if event != Job.NOTIFY_RUN_DONE:
            return

        # TODO: this should only start runs on the same node if this is an
        # all_nodes job, but that is currently not possible
        queued_run = self.job.runs.get_first_queued()
        if queued_run:
            reactor.callLater(0, self.run_job, queued_run, run_queued=True)

        if self.job.scheduler.schedule_on_complete:
            self.schedule()
    handler = handle_job_events

    def get_runs_to_schedule(self, ignore_last_run_time=False):
        """Build and return the runs to schedule."""
        if self.job.runs.has_pending:
            log.info("%s has pending runs, can't schedule more." % self.job)
            return []

        if ignore_last_run_time:
            last_run_time = None
        else:
            last_run = self.job.runs.get_newest(include_manual=False)
            last_run_time = last_run.run_time if last_run else None
        next_run_time = self.job.scheduler.next_run_time(last_run_time)
        return self.job.build_new_runs(next_run_time)


import collections
import logging

from twisted.internet import reactor

from tron import command_context
from tron import event
from tron import node
from tron.actioncommand import ActionCommand
from tron.utils import state, observer
from tron.utils import timeutils


log = logging.getLogger(__name__)

MIN_MONITOR_HANG_TIME = 10


class Error(Exception):
    """Generic service error"""


class InvalidStateError(Error):
    """Invalid state error"""


class ServiceInstance(observer.Observer):
    class ServiceInstanceState(state.NamedEventState):
        """Event state subclass for service instances"""

    STATE_DOWN = ServiceInstanceState("down")

    STATE_UP = ServiceInstanceState("up")

    STATE_FAILED = ServiceInstanceState("failed",
                                        stop=STATE_DOWN,
                                        up=STATE_UP)

    STATE_STOPPING = ServiceInstanceState("stopping",
                                          down=STATE_DOWN)

    STATE_MONITORING = ServiceInstanceState("monitoring",
                                            down=STATE_FAILED,
                                            stop=STATE_STOPPING,
                                            up=STATE_UP)

    STATE_STARTING = ServiceInstanceState("starting",
                                          down=STATE_FAILED,
                                          monitor=STATE_MONITORING,
                                          stop=STATE_STOPPING)

    STATE_UNKNOWN = ServiceInstanceState("unknown",
                                         monitor=STATE_MONITORING)

    STATE_MONITORING['monitor_fail'] = STATE_UNKNOWN

    STATE_UP['stop'] = STATE_STOPPING

    STATE_UP['monitor'] = STATE_MONITORING

    STATE_DOWN['start'] = STATE_STARTING

    def __init__(self, service, node, instance_number):
        self.service = service
        self.instance_number = instance_number
        self.node = node

        self.id = "%s.%s" % (service.name, self.instance_number)

        self.machine = state.StateMachine(
                ServiceInstance.STATE_DOWN, delegate=self)

        self.context = command_context.CommandContext(self, service.context)

        self.monitor_action = None
        self.start_action   = None
        self.stop_action    = None

        # Store the Twisted delayed call objects here for later cancellation
        self._monitor_delayed_call = None
        self._hanging_monitor_check_delayed_call = None

    @property
    def state(self):
        return self.machine.state

    @property
    def attach(self):
        return self.machine.attach

    @property
    def pid_file(self):
        if self.service.pid_file_template:
            try:
                return self.service.pid_file_template % self.context
            except KeyError:
                log.error("Failed to render pid file template: %r",
                          self.service.pid_file_template)
        else:
            log.warning("No pid_file configured for service %s",
                        self.service.name)

        return None

    @property
    def command(self):
        try:
            return self.service.command % self.context
        except KeyError:
            log.error("Failed to render service command for service %s: %s",
                      self.service.name, self.service.command)

        return None

    def _queue_monitor(self):
        self.monitor_action = None
        if self.service.monitor_interval > 0:
            self._monitor_delayed_call = reactor.callLater(
                self.service.monitor_interval, self._run_monitor)

    def _queue_monitor_hang_check(self):
        """Since our monitor cycle is controlled by monitors actually
        completing, we need a check to ensure the monitor doesn't hang.

        We aren't going to make this monitor interval configurable right now,
        but just peg it to a factor of the interval.
        """

        current_action = self.monitor_action
        hang_monitor_duration = max((self.service.monitor_interval or 0) * 0.8,
                                    MIN_MONITOR_HANG_TIME)

        self._hanging_monitor_check_delayed_call = reactor.callLater(
            hang_monitor_duration,
            lambda: self._monitor_hang_check(current_action))

    def _run_monitor(self):
        self._monitor_delayed_call = None

        if self.monitor_action:
            log.warning("Monitor action already exists, old callLater ?")
            return

        self.machine.transition("monitor")
        pid_file = self.pid_file

        if pid_file is None:
            # If our pid file doesn't exist or failed to be generated, we
            # can't really monitor
            self._monitor_complete_failstart()
            return

        monitor_command = "cat %(pid_file)s | xargs kill -0" % self.context
        log.debug("Executing '%s' on %s for %s", monitor_command,
                  self.node.hostname, self.id)
        self.monitor_action = ActionCommand("%s.monitor" % self.id,
                                                   monitor_command)
        self.watch(self.monitor_action)

        try:
            self.node.run(self.monitor_action)
        except node.Error, e:
            log.error("Failed to run monitor: %r", e)
            return

        self._queue_monitor_hang_check()

    def _monitor_hang_check(self, action):
        self._hanging_monitor_check_delayed_call = None
        if self.monitor_action is action:
            log.warning("Monitor for %s is still running", self.id)
            self.machine.transition("monitor_fail")
            self._queue_monitor_hang_check()

    def _monitor_complete_callback(self):
        """Callback when our monitor has completed"""
        if not self.monitor_action:
            # This actually happened. I suspect it was a cascading failure
            # caused by a crash somewhere else leaving us in an inconsistent
            # state, but perhaps there is a reasonable explanation. Either way,
            # we don't really care about this monitor anymore.
            log.warning("Monitor for %s complete, but we don't see to care...",
                        self.id)
            return

        self.last_check = timeutils.current_time()
        log.debug("Monitor callback with exit %r",
                  self.monitor_action.exit_status)
        if self.monitor_action.exit_status != 0:
            self.machine.transition("down")
        else:
            self.machine.transition("up")
            self._queue_monitor()

        self.monitor_action = None

    def _monitor_complete_failstart(self):
        """Callback when our monitor failed to even start"""
        self.machine.transition("monitor_fail")
        self._queue_monitor()

        self.monitor_action = None

    def start(self):
        if self.machine.state != self.STATE_DOWN:
            raise InvalidStateError("Instance must be marked DOWN to start")

        self.machine.transition("start")

        command = self.command
        if command is None:
            self._start_complete_failstart()
            return

        self.start_action = ActionCommand("%s.start" % self.id, command)
        self.watch(self.start_action)

        try:
            self.node.run(self.start_action)
        except node.Error, e:
            log.warning("Failed to start %s: %r", self.id, e)

    def handler(self, observable, event):
        # TODO: this no longer requires setting the start action as a field
        # TODO: This can be easily cleaned up

        if observable == self.monitor_action:
            # We use exiting instead of complete because all we really need is
            # the exit status
            if event == ActionCommand.EXITING:
                return self._monitor_complete_callback()
            if event == ActionCommand.FAILSTART:
                return self._monitor_complete_failstart()

        elif observable == self.start_action:
            if event == ActionCommand.EXITING:
                return self._start_complete_callback()
            if event == ActionCommand.FAILSTART:
                return self._start_complete_failstart()

        elif observable == self.stop_action:
            if event == ActionCommand.COMPLETE:
                return self._stop_complete_callback()
            if event == ActionCommand.FAILSTART:
                return self._stop_complete_failstart()

    def _start_complete_callback(self):
        if self.start_action.exit_status != 0:
            self.machine.transition("down")
        elif self.machine.state == self.STATE_STOPPING:
            # Someone tried to stop us while we were just getting going. Go
            # ahead and kick of the kill operation now that we're up.
            if self.stop_action:
                log.warning("Stopping %s while stop already in progress",
                            self.id)
            else:
                self.kill_instance()
        else:
            log.info("Start for %s complete, checking monitor", self.id)
            self._queue_monitor()

        self.start_action = None

    def _start_complete_failstart(self):
        log.warning("Failed to start service %s (%s)",
                    self.id, self.node.hostname)

        # We may have failed but long since not mattered
        if None in (self.machine, self.start_action):
            return

        self.machine.transition("down")
        self.start_action = None

    def stop(self):
        self.machine.transition("stop")

        if self.machine.state == self.STATE_STOPPING:
            self.kill_instance()

    def kill_instance(self):
        assert self.pid_file, self.pid_file

        kill_command = "cat %(pid_file)s | xargs kill" % self.context

        self.stop_action = ActionCommand("%s.stop" % self.id,
                                                kill_command)
        self.watch(self.stop_action)
        try:
            self.node.run(self.stop_action)
        except node.Error, e:
            log.warning("Failed to kill instance %s: %r", self.id, e)

    def _stop_complete_callback(self):
        if self.stop_action.exit_status != 0:
            log.error("Failed to stop service instance %s: Exit %r", self.id,
                      self.stop_action.exit_status)

        self._queue_monitor()
        self.stop_action = None

    def _stop_complete_failstart(self):
        log.warning("Failed to start kill command for %s", self.id)
        self._queue_monitor()

    def zap(self):
        self.machine.transition("stop")
        self.machine.transition("down")

        # Kill the monitors so they don't put us back in the UP state
        if self._monitor_delayed_call is not None:
            self._monitor_delayed_call.cancel()
            self._monitor_delayed_call = None

        if self._hanging_monitor_check_delayed_call is not None:
            self._hanging_monitor_check_delayed_call.cancel()
            self._hanging_monitor_check_delayed_call = None

    @property
    def state_data(self):
        """This data is used to serialize the state of this service instance."""
        return {
            'node': self.node.hostname,
            'instance_number': self.instance_number,
            'state': str(self.state),
        }

    def __str__(self):
        return "SERVICE:%s" % self.id


class Service(observer.Observable, observer.Observer):
    # For comparing equality, we check these fields
    COMPARE_ATTRIBUTES = ['name', 'command', 'node_pool', 'count',
                          'monitor_interval', 'pid_file_template']

    class ServiceState(state.NamedEventState):
        """Named event state subclass for services"""

    STATE_DOWN = ServiceState("down")

    STATE_UP = ServiceState("up")

    STATE_DEGRADED = ServiceState("degraded")

    STATE_STOPPING = ServiceState("stopping", all_down=STATE_DOWN)

    STATE_FAILED = ServiceState("failed")

    STATE_STARTING = ServiceState("starting",
                                  all_up=STATE_UP,
                                  failed=STATE_DEGRADED,
                                  stop=STATE_STOPPING)

    STATE_DOWN['start'] = STATE_STARTING

    STATE_DEGRADED.update(dict(stop=STATE_STOPPING,
                               all_up=STATE_UP,
                               all_failed=STATE_FAILED))

    STATE_FAILED.update(dict(stop=STATE_STOPPING,
                             up=STATE_DEGRADED,
                             start=STATE_STARTING))

    STATE_UP.update(dict(stop=STATE_STOPPING,
                         failed=STATE_DEGRADED,
                         down=STATE_DEGRADED))

    def __init__(self, name=None, command=None, node_pool=None, context=None,
                 event_recorder=None, monitor_interval=None,
                 restart_interval=None, pid_file_template=None, count=0):
        super(Service, self).__init__()
        self.name = name
        self.command = command
        self.node_pool = node_pool
        self.count = count
        self.monitor_interval = monitor_interval
        self.restart_interval = restart_interval
        self._restart_timer = None

        self.machine = state.StateMachine(Service.STATE_DOWN, delegate=self)

        self.pid_file_template = pid_file_template

        self.context = None
        if context is not None:
            self.set_context(context)

        self.instances = []

        self.event_recorder = event.EventRecorder(self, parent=event_recorder)
        # TODO: this is a little weird, should get cleaned up
        self.listen(True, self)

    @classmethod
    def from_config(cls, srv_config, node_pools):
        return cls(
            name=srv_config.name,
            node_pool=node_pools[srv_config.node] if srv_config.node else None,
            monitor_interval=srv_config.monitor_interval,
            restart_interval=srv_config.restart_interval,
            pid_file_template=srv_config.pid_file,
            command=srv_config.command,
            count=srv_config.count
        )

    @property
    def state(self):
        return self.machine.state

    @property
    def listen(self):
        return self.machine.attach

    @property
    def is_started(self):
        """Indicate if the service has been started/initialized

        For now we're going to decide this if we have instances or not. It
        doesn't really correspond well to a "state", but it might at some point
        need to be some sort of enable/disable thing.
        """
        return len(self.instances) > 0

    def set_context(self, context):
        self.context = command_context.CommandContext(self, context)

    def _record_state_changes(self):
        # If we no longer have a state machine, that means we no longer matter
        if self.machine is None:
            return

        if self.machine.state in (self.STATE_FAILED, self.STATE_DEGRADED):
            func = self.event_recorder.emit_critical
        elif self.machine.state in (self.STATE_UP, self.STATE_DOWN):
            func = self.event_recorder.emit_ok
        else:
            func = self.event_recorder.emit_info

        if func:
            func(str(self.machine.state))

    def _clear_failed_instances(self):
        """Remove and cleanup any instances that are no longer with us"""
        self.instances = [inst for inst in self.instances
                          if inst.state != ServiceInstance.STATE_FAILED]

    def _restart_after_failure(self):
        if self._restart_timer is None or self.machine is None:
            return

        if self.state in (self.STATE_DEGRADED, self.STATE_FAILED):
            log.info("Restarting failed instances for service %s", self.name)
            self.start()
        else:
            self._restart_timer = None

    def start(self):
        # Clear out the restart timer, just to make sure we don't get any
        # extraneous starts
        self._restart_timer = None

        # Start can really mean restart any failed or down instances.
        # So first off, clear out any old instances that are of no use to us
        # anymore
        self._clear_failed_instances()

        # Build all the new instances well need
        needed_instances_count = self.count - len(self.instances)
        if needed_instances_count > 0:
            for _ in range(0, needed_instances_count):
                self.build_instance()

        self.machine.transition("start")

    def stop(self):
        self.machine.transition("stop")

        for service_instance in self.instances:
            service_instance.stop()

        # Just in case we somehow ended up stuck with no instances, double
        # check here for stop complete.
        if self.state == self.STATE_STOPPING and not self.instances:
            self.machine.transition("all_down")

    def zap(self):
        self.machine.transition("stop")

        for service_instance in self.instances:
            service_instance.zap()

        if self.state == self.STATE_STOPPING and not self.instances:
            self.machine.transition("all_down")

    def _create_instance(self, node, instance_number):
        service_instance = ServiceInstance(self, node, instance_number)
        self.instances.append(service_instance)
        self.instances.sort(key=lambda i: i.instance_number)
        self.watch(service_instance)
        return service_instance

    def _find_unused_instance_number(self):
        available_instance_numbers = (set(range(0, self.count)) -
                                      set(instance.instance_number
                                          for instance in self.instances))

        if len(available_instance_numbers) == 0:
            return None

        return min(available_instance_numbers)

    def build_instance(self):
        node = self.node_pool.next_round_robin()

        instance_number = self._find_unused_instance_number()

        if instance_number is None:
            log.error("Can't build a new service instance for %r. %d instances"
                      " in use. Maybe try again later ?",
                      self.name, self.count)
            return None

        service_instance = self._create_instance(node, instance_number)

        # No reason not to start this guy right away, we don't keep 'down'
        # instances around really.
        service_instance.start()

        return service_instance

    def _instance_change(self):
        """Handle any changes to our service's instances

        This is the state change callback handler for all our instances.
        Anytime an instance changes, we need to re-evaluate our own current
        state.
        """
        #TODO: this can be improved now that we have a reference to
        # the instance that changed

        # Remove any downed instances
        self.instances = [inst for inst in self.instances
                          if inst.state != inst.STATE_DOWN]

        # Now we can make some inferences about state changes based on our
        # instances
        if not self.instances:
            self.machine.transition("all_down")

        elif any([instance.state == ServiceInstance.STATE_FAILED
                  for instance in self.instances]):
            self.machine.transition("failed")

            if all([instance.state == ServiceInstance.STATE_FAILED
                    for instance in self.instances]):
                self.machine.transition("all_failed")

        elif len(self.instances) < self.count:
            log.info("Only found %d instances rather than %d",
                     len(self.instances), self.count)
            self.machine.transition("down")

        elif all(instance.state == ServiceInstance.STATE_UP
                  for instance in self.instances):
            self.machine.transition("all_up")

        if self.machine.state in (Service.STATE_DEGRADED,
                                  Service.STATE_FAILED):
            # Start a restart timer if configure
            if self.restart_interval is not None and not self._restart_timer:
                self._restart_timer = reactor.callLater(
                    self.restart_interval, self._restart_after_failure)

    def handler(self, observable, _event):
        if observable == self:
            return self._record_state_changes()

        self._instance_change()

    # TODO: clean this up
    def absorb_previous(self, prev_service):
        # Some changes we need to worry about:
        # * Changing instance counts
        # * Changing the command
        # * Changing the node pool
        # * Changes to the context ?
        # * Restart counts for downed services ?
        assert self.node_pool, "Missing node pool for %s" % self.name
        removed_instances = 0

        rebuild_all_instances = any([
            self.command != prev_service.command,
            self.pid_file_template != prev_service.pid_file_template])

        # Since we are inheriting all the existing instances, it's safe to also
        # inherit the previous state machine as well.
        self.machine = prev_service.machine

        # To permanently disable the older service, remove it's machine.
        prev_service.machine = None

        # Copy over all the old instances
        self.instances += prev_service.instances
        for service_instance in prev_service.instances:
            service_instance.machine.clear_observers()
            self.watch(service_instance)

            if rebuild_all_instances:
                # For some configuration changes, we'll just stop all the
                # previous instances. When those services stop, we should be in
                # a degraded mode, triggering a restart of the newer generation
                # of instances.
                service_instance.stop()
                removed_instances += 1

        prev_service.instances = []

        self.instances.sort(key=lambda i: i.instance_number)

        current_instances = [i for i in self.instances if i.state not in
                             (ServiceInstance.STATE_STOPPING,
                              ServiceInstance.STATE_DOWN,
                              ServiceInstance.STATE_FAILED)]

        # Now that we've inherited some instances, let's trigger an update to
        # our state machine.
        self._instance_change()

        # We have special handling for node pool changes. This would cover the
        # case of removing (or subsituting) a node in a pool which would
        # require rebalancing services.

        if self.node_pool != prev_service.node_pool:
            # How many instances per node should we have ?
            optimal_instances_per_node = self.count / len(self.node_pool.nodes)
            instance_count_per_node = collections.defaultdict(int)

            for service_instance in current_instances:
                # First we'll stop any instances on nodes that are no longer
                # part of our pool
                try:
                    hostname = service_instance.node.hostname
                    service_instance.node = self.node_pool[hostname]
                except KeyError:
                    log.info("Stopping instance %r because it's not on a"
                             " current node (%r)",
                             service_instance.id,
                             service_instance.node.hostname)
                    service_instance.stop()
                    removed_instances += 1
                    continue

                instance_count_per_node[service_instance.node] += 1
                if (instance_count_per_node[service_instance.node] >
                        optimal_instances_per_node):
                    log.info("Stopping instance %r because node %s has too"
                             " many instances",
                             service_instance.id,
                             service_instance.node.hostname)
                    service_instance.stop()
                    removed_instances += 1
                    continue

        current_instances = [i for i in self.instances if i.state not in
                             (ServiceInstance.STATE_STOPPING,
                              ServiceInstance.STATE_DOWN,
                              ServiceInstance.STATE_FAILED)]

        count_to_remove = ((len(self.instances) - removed_instances) -
                           self.count)
        if count_to_remove > 0:
            instances_to_remove = current_instances[-count_to_remove:]
            for service_instance in instances_to_remove:
                service_instance.stop()
                removed_instances += 1

        # Now make adjustments to how many there are
        while len(self.instances) < self.count:
            self.build_instance()

        self.event_recorder.emit_notice("reconfigured")

    @property
    def state_data(self):
        """This data is used to serialize the state of this service."""
        data = {
            'state': str(self.machine.state),
            'instances': [instance.state_data for instance in self.instances]
        }
        return data

    def restore(self, data):
        """Restore state of this service from datafile"""
        # The state of a service is more easier than for jobs. There are just a
        # few things we want to guarantee:
        #  1. If service instances are up, they can continue to be up. We'll
        #     just start monitoring from where we left off.
        #  2. Failures are maintained and have to be cleared.

        # Start our machine from where it left off
        self.machine.state = state.named_event_by_name(Service.STATE_DOWN,
                                                       data['state'])

        if self.machine.state in (Service.STATE_DOWN, Service.STATE_FAILED):
            self.event_recorder.emit_info("restored")
            return

        # Restore all the instances
        # We're going to just indicate they are up and start a monitor
        for instance in data['instances']:
            try:
                node = self.node_pool[instance['node']]
            except KeyError:
                log.error("Failed to find node %s in pool for %s",
                          instance['node'],
                          self.name)
                continue

            service_instance = self._create_instance(
                node, instance['instance_number'])
            service_instance.machine.state = ServiceInstance.STATE_MONITORING
            service_instance._run_monitor()

        self.instances.sort(key=lambda i: i.instance_number)
        self.event_recorder.emit_info("restored")

    def __eq__(self, other):
        if other is None or not isinstance(other, Service):
            return False

        for attr_name in self.COMPARE_ATTRIBUTES:
            if getattr(self, attr_name) != getattr(other, attr_name):
                return False

        return True

    def __str__(self):
        return "SERVICE:%s" % self.name



"""
Tron schedulers

 A scheduler has a simple interface.

 class Scheduler(object):

    schedule_on_complete = <bool>

    def next_run_time(self, last_run_time):
        <returns datetime>


 next_run_time() should return a datetime which is the time the next job run
 will be run.

 schedule_on_complete is a bool that identifies if this scheduler should have
 jobs scheduled with the start_time of the previous run (False), or the
 end time of the previous run (False).
"""
import logging

from pytz import AmbiguousTimeError, NonExistentTimeError
from tron.config import schedule_parse

from tron.utils import groctimespecification
from tron.utils import timeutils


log = logging.getLogger(__name__)


def scheduler_from_config(config, time_zone):
    """A factory for creating a scheduler from a configuration object."""
    if isinstance(config, schedule_parse.ConfigConstantScheduler):
        return ConstantScheduler()

    if isinstance(config, schedule_parse.ConfigIntervalScheduler):
        return IntervalScheduler(interval=config.timedelta)

    if isinstance(config, schedule_parse.ConfigDailyScheduler):
        return DailyScheduler(
            time_zone=time_zone,
            timestr=config.timestr,
            ordinals=config.ordinals,
            monthdays=config.monthdays,
            months=config.months,
            weekdays=config.weekdays
        )


class ConstantScheduler(object):
    """The constant scheduler schedules a new job immediately."""
    schedule_on_complete = True

    def next_run_time(self, _):
        return timeutils.current_time()

    def __str__(self):
        return "CONSTANT"

    def __eq__(self, other):
        return isinstance(other, ConstantScheduler)

    def __ne__(self, other):
        return not self == other


class DailyScheduler(object):
    """Wrapper around SpecificTimeSpecification in the Google App Engine cron
    library
    """
    schedule_on_complete = False

    def __init__(self, ordinals=None, weekdays=None, months=None,
                 monthdays=None, timestr=None, time_zone=None,
                 string_repr=None):
        """Parameters:
          timestr     - the time of day to run, as 'HH:MM'
          ordinals    - first, second, third &c, as a set of integers in 1..5 to
                        be used with "1st <weekday>", etc.
          monthdays   - set of integers to be used with "<month> 3rd", etc.
          months      - the months that this should run, as a set of integers in
                        1..12
          weekdays    - the days of the week that this should run, as a set of
                        integers, 0=Sunday, 6=Saturday
          timezone    - the optional timezone as a string for this specification.
                        Defaults to UTC - valid entries are things like
                        Australia/Victoria or PST8PDT.
          string_repr - Original string representation this was parsed from,
                        if applicable
        """
        self.ordinals               = ordinals
        self.weekdays               = weekdays
        self.months                 = months
        self.monthdays              = monthdays
        self.timestr                = timestr or '00:00'
        self.time_zone              = time_zone
        self.string_repr            = string_repr
        self._time_spec             = None

    @property
    def time_spec(self):
        if self._time_spec is None:
            # calendar module has 0=Monday
            # groc has 0=Sunday
            if self.weekdays:
                groc_weekdays = set((x + 1) % 7 for x in self.weekdays)
            else:
                groc_weekdays = None
            self._time_spec = groctimespecification.SpecificTimeSpecification(
                ordinals=self.ordinals,
                weekdays=groc_weekdays,
                months=self.months,
                monthdays=self.monthdays,
                timestr=self.timestr,
                timezone=self.time_zone.zone if self.time_zone else None,
            )
        return self._time_spec

    def next_run_time(self, start_time):
        """Find the next time to run."""
        if not start_time:
            start_time = timeutils.current_time()
        elif self.time_zone:
            try:
                start_time = self.time_zone.localize(start_time, is_dst=None)
            except AmbiguousTimeError:
                # We are in the infamous 1 AM block which happens twice on
                # fall-back. Pretend like it's the first time, every time.
                start_time = self.time_zone.localize(start_time, is_dst=True)
            except NonExistentTimeError:
                # We are in the infamous 2:xx AM block which does not
                # exist. Pretend like it's the later time, every time.
                start_time = self.time_zone.localize(start_time, is_dst=True)

        return self.time_spec.GetMatch(start_time)

    def __str__(self):
        # Backward compatible string representation which also happens to be
        # user-friendly
        if self.string_repr is None:
            return 'DAILY'
        else:
            return self.string_repr

    def __eq__(self, other):
        return isinstance(other, DailyScheduler) and \
           all(getattr(self, attr) == getattr(other, attr)
               for attr in ('ordinals',
                            'weekdays',
                            'months',
                            'monthdays',
                            'timestr',
                            'time_zone'))

    def __ne__(self, other):
        return not self == other


class IntervalScheduler(object):
    """The interval scheduler runs a job (to success) based on a configured
    interval.
    """
    schedule_on_complete = False

    def __init__(self, interval=None):
        self.interval = interval

    def next_run_time(self, last_run_time):
        last_run_time = last_run_time or timeutils.current_time()
        return last_run_time + self.interval

    def __str__(self):
        return "INTERVAL:%s" % self.interval

    def __eq__(self, other):
        return (isinstance(other, IntervalScheduler) and
                self.interval == other.interval)

    def __ne__(self, other):
        return not self == other



"""
Common code for command line utilities (see bin/)
"""
from __future__ import with_statement
import logging
import os
import os.path
import sys

import yaml


GLOBAL_CONFIG_FILE_NAME = os.environ.get('TRON_CONFIG') or "/etc/tron/tron.yaml"
CONFIG_FILE_NAME = "~/.tron"

DEFAULT_HOST = 'localhost'
DEFAULT_PORT = 8089

DEFAULT_CONFIG = {
    'server':           "http://%s:%d" % (DEFAULT_HOST, DEFAULT_PORT),
    'display_color':    False,
}

log = logging.getLogger("tron.cmd")


def load_config(options):
    """Attempt to load a user specific configuration or a global config file
    and set any unset options based on values from the config. Finally fallback
    to DEFAULT_CONFIG for those settings.
    """
    config = {}
    config_file_list = [CONFIG_FILE_NAME, GLOBAL_CONFIG_FILE_NAME]
    for config_file in config_file_list:
        file_name = os.path.expanduser(config_file)
        if os.access(file_name, os.R_OK):
            try:
                with open(file_name, "r") as config_file:
                    config = yaml.load(config_file)
                break
            except IOError, e:
                log.error("Failure loading config file: %r", e)

    else:
        log.debug("Could not find a config in: %s." % ", ".join(config_file_list))

    for opt_name in DEFAULT_CONFIG.keys():
        if not hasattr(options, opt_name):
            continue

        if getattr(options, opt_name) is not None:
            continue

        default_value = DEFAULT_CONFIG[opt_name]
        setattr(options, opt_name, config.get(opt_name, default_value))


def save_config(options):
    file_name = os.path.expanduser(CONFIG_FILE_NAME)

    try:
        with open(file_name, "r") as config_file:
            config = yaml.load(config_file)
    except IOError:
        log.info("Failed to locate an existing config file %s" % file_name)
        config = None

    config = config or {}
    for opt_name in DEFAULT_CONFIG.keys():
        if not hasattr(options, opt_name):
            continue
        config[opt_name] = getattr(options, opt_name)

    with open(file_name, "w") as config_file:
        yaml.dump(config, config_file)


def setup_logging(options):
    level = logging.INFO if options.verbose else logging.WARNING

    logging.basicConfig(
        level=level,
        format='%(name)s %(levelname)s %(message)s',
        stream=sys.stdout
    )



import logging
import itertools
import random

from twisted.internet import protocol, defer, reactor
from twisted.python import failure

from tron import ssh
from tron.utils import twistedutils


log = logging.getLogger(__name__)

# We should also only wait a certain amount of time for a connection to be
# established.
CONNECT_TIMEOUT = 30

IDLE_CONNECTION_TIMEOUT = 3600

# We should also only wait a certain amount of time for a new channel to be
# established when we already have an open connection.  This timeout will
# usually get triggered prior to even a TCP timeout, so essentially it's our
# shortcut to discovering the connection died.
RUN_START_TIMEOUT = 20

# Love to run this, but we need to finish connecting to our node first
RUN_STATE_CONNECTING = 0

# We are connected and trying to open a channel to exec the process
RUN_STATE_STARTING = 5

# Process has been exec'ed, just waiting for it to exit
RUN_STATE_RUNNING = 10

# Process has exited
RUN_STATE_COMPLETE = 100


class Error(Exception):
    pass


class ConnectError(Error):
    """There was a problem connecting, run was never started"""
    pass


class ResultError(Error):
    """There was a problem retrieving the result from this run

    We did try to execute the command, but we don't know if it succeeded or
    failed.
    """
    pass


class RunState(object):
    def __init__(self, run):
        self.run = run
        self.state = RUN_STATE_CONNECTING
        self.deferred = defer.Deferred()
        self.channel = None


class NodePoolStore(dict):
    """A Singleton to store Node and NodePool objects."""

    _instance = None

    def __init__(self):
        if self._instance is not None:
            raise ValueError("NodePoolStore is already instantiated.")
        super(NodePoolStore, self).__init__()

    @classmethod
    def get_instance(cls):
        if cls._instance is None:
            cls._instance = cls()
        return cls._instance

    def put(self, node):
        self[node.name] = node

    def update(self, nodes):
        super(NodePoolStore, self).update((node.name, node) for node in nodes)


class NodePool(object):
    """A pool of Node objects."""
    def __init__(self, nodes, name=None):
        self.nodes = nodes
        self.name = name or '_'.join(n.name for n in nodes)
        self.iter = itertools.cycle(self.nodes)

    @classmethod
    def from_config(cls, node_pool_config):
        nodes = NodePoolStore.get_instance()
        return cls(
            name=node_pool_config.name,
            nodes=[nodes[n] for n in node_pool_config.nodes]
        )

    def __eq__(self, other):
        return isinstance(other, NodePool) and self.nodes == other.nodes

    def __ne__(self, other):
        return not self == other

    def next(self):
        """Return a random node from the pool."""
        return random.choice(self.nodes)

    def next_round_robin(self):
        """Return the next node cycling in a consistent order."""
        return self.iter.next()

    def __getitem__(self, value):
        for node in self.nodes:
            if node.hostname == value:
                return node
        raise KeyError(value)

    def repr_data(self):
        """Returns a dict which is an external view of this object."""
        return {
            'name':         self.name,
            'nodes':        [n.repr_data() for n in self.nodes]
        }

class Node(object):
    """A node is tron's interface to communicating with an actual machine.
    This class also supports the NodePool interface and can be used
    directly as a NodePool of 1 Node.
    """

    def __init__(self, hostname=None, name=None, ssh_options=None):
        # Host we are to connect to
        self.hostname = hostname

        # Identifier for UI
        self.name = name or hostname

        if not ssh_options or not hostname:
            raise ValueError('Must specify hostname and ssh_options')

        self.conch_options = ssh_options

        # The SSH connection we use to open channels on. If present, means we
        # are connected.
        self.connection = None

        # If present, means we are trying to connect
        self.connection_defer = None

        # Map of run id to instance of RunState
        self.run_states = {}

        self.idle_timeout = None
        self.idle_timer = None

    @classmethod
    def from_config(cls, node_config, ssh_options):
        return cls(
            hostname=node_config.hostname,
            name=node_config.name,
            ssh_options=ssh_options
        )

    def next(self):
        """Required to support the NodePool interface."""
        return self

    def next_round_robin(self):
        """Required to support the NodePool interface."""
        return self

    @property
    def nodes(self):
        """Required to support the NodePool interface."""
        return [self]

    def __getitem__(self, value):
        """Required to support the NodePool interface."""
        if self.hostname == value:
            return self
        raise KeyError(value)

    def __cmp__(self, other):
        if not isinstance(other, self.__class__):
            return -1

        return cmp(self.hostname, other.hostname)

    def _determine_fudge_factor(self):
        """We want to introduce some amount of delay to node exec commands

        We see issues where a service may have many instances, and they all
        start at once, and their monitor steps are blocked for ever. This
        is bad.
        """
        outstanding_runs = len(self.run_states)
        fudge_factor = max(0.0, outstanding_runs - 4)
        return random.random() * float(fudge_factor)

    # TODO: Test
    def submit_command(self, command):
        """Submit an ActionCommand to be run on this node. Optionally provide
        an error callback which will be called on error.
        """
        deferred = self.run(command)
        deferred.addErrback(command.handle_errback)
        return deferred

    def run(self, run):
        """Execute the specified run

        A run consists of a very specific set of interfaces which allow us to
        execute a command on this remote machine and return results.
        """
        log.info("Running %s on %s", run.id, self.hostname)

        # When this run completes, for good or bad, we'll inform the caller by
        # calling 'succeed' or 'fail' on the run Since the definined interface
        # is on these specific callbacks, we won't bother returning the
        # deferred here. This allows the caller to not really care about
        # twisted specific stuff at all, all it needs to know is that one of
        # those functions will eventually be called back

        if run.id in self.run_states:
            raise Error("Run %s already running !?!", run.id)

        if self.idle_timer is not None:
            self.idle_timer.cancel()
            self.idle_timer = None

        fudge_factor = self._determine_fudge_factor()

        self.run_states[run.id] = RunState(run)

        if fudge_factor == 0.0:
            self._do_run(run)
        else:
            log.info("Delaying execution of %s for %.2f secs",
                     run.id, fudge_factor)
            reactor.callLater(fudge_factor, lambda: self._do_run(run))

        # We return the deferred here, but really we're trying to keep the rest
        # of the world from getting too involved with twisted.
        return self.run_states[run.id].deferred

    def _do_run(self, run):
        """Finish starting to execute a run

        This step may have been delayed.
        """

        # Now let's see if we need to start this off by establishing a
        # connection or if we are already connected
        if self.connection is None:
            self._connect_then_run(run)
        else:
            self._open_channel(run)

    def _cleanup(self, run):
        self.run_states[run.id].channel = None
        del self.run_states[run.id]

        if not self.run_states:
            self.idle_timer = reactor.callLater(IDLE_CONNECTION_TIMEOUT,
                                                self._connection_idle_timeout)

    def _connection_idle_timeout(self):
        if self.connection:
            log.info("Connection to %s idle for %d secs. Closing.",
                     self.hostname, IDLE_CONNECTION_TIMEOUT)
            self.connection.transport.loseConnection()

        self.idle_timer = None

    def _fail_run(self, run, result):
        """Indicate the run has failed, and cleanup state"""
        log.debug("Run %s has failed", run.id)
        if run.id not in self.run_states:
            log.warning("Run %s no longer tracked (_fail_run)", run.id)
            return

        # Add a dummy errback handler to prevent Unhandled error messages.
        # Unless somone is explicitly caring about this defer the error will
        # have been reported elsewhere.
        self.run_states[run.id].deferred.addErrback(lambda failure: None)

        cb = self.run_states[run.id].deferred.errback

        self._cleanup(run)

        log.info("Calling fail_run callbacks")
        run.exited(None)
        cb(result)

    def _connect_then_run(self, run):
        # Have we started the connection process ?
        if self.connection_defer is None:
            self.connection_defer = self._connect()

        def call_open_channel(arg):
            self._open_channel(run)
            return arg

        def connect_fail(result):
            log.warning("Cannot run %s, Failed to connect to %s",
                        run.id, self.hostname)
            self.connection_defer = None
            self._fail_run(run, failure.Failure(
                exc_value=ConnectError("Connection to %s failed" %
                                       self.hostname)))

        self.connection_defer.addCallback(call_open_channel)
        self.connection_defer.addErrback(connect_fail)

    def _service_stopped(self, connection):
        """Called when the SSH service has disconnected fully.

        We should be in a state where we know there are no runs in progress
        because all the SSH channels should have disconnected them.
        """
        assert self.connection is connection
        self.connection = None

        log.info("Service to %s stopped", self.hostname)

        for run_id, run in self.run_states.iteritems():
            if run.state == RUN_STATE_CONNECTING:
                # Now we can trigger a reconnect and re-start any waiting runs.
                self._connect_then_run(run)
            elif run.state == RUN_STATE_RUNNING:
                self._fail_run(run, None)
            elif run.state == RUN_STATE_STARTING:
                if run.channel and run.channel.start_defer is not None:

                    # This means our run IS still waiting to start. There
                    # should be an outstanding timeout sitting on this guy as
                    # well. We'll just short circut it.
                    twistedutils.defer_timeout(run.channel.start_defer, 0)
                else:
                    # Doesn't seem like this should ever happen.
                    log.warning("Run %r caught in starting state, but"
                                " start_defer is over.", run_id)
                    self._fail_run(run, None)
            else:
                # Service ended. The open channels should know how to handle
                # this (and cleanup) themselves, so if there should not be any
                # runs except those waiting to connect
                raise Error("Run %s in state %s when service stopped",
                            run_id, run.state)

    def _connect(self):
        # This is complicated because we have to deal with a few different
        # steps before our connection is really available for us:
        #  1. Transport is created (our client creator does this)
        #  2. Our transport is secure, and we can create our connection
        #  3. The connection service is started, so we can use it

        client_creator = protocol.ClientCreator(reactor, ssh.ClientTransport,
                                                options=self.conch_options)
        create_defer = client_creator.connectTCP(self.hostname, 22)

        # We're going to create a deferred, returned to the caller, that will
        # be called back when we have an established, secure connection ready
        # for opening channels. The value will be this instance of node.
        connect_defer = defer.Deferred()
        twistedutils.defer_timeout(connect_defer, CONNECT_TIMEOUT)

        def on_service_started(connection):
            # Booyah, time to start doing stuff
            self.connection = connection
            self.connection_defer = None

            connect_defer.callback(self)
            return connection

        def on_connection_secure(connection):
            # We have a connection, but it might not be fully ready....
            connection.service_start_defer = defer.Deferred()
            connection.service_stop_defer = defer.Deferred()

            connection.service_start_defer.addCallback(on_service_started)
            connection.service_stop_defer.addCallback(self._service_stopped)
            return connection

        def on_transport_create(transport):
            transport.connection_defer = defer.Deferred()
            transport.connection_defer.addCallback(on_connection_secure)
            return transport

        def on_transport_fail(fail):
            log.warning("Cannot connect to %s", self.hostname)

        create_defer.addCallback(on_transport_create)
        create_defer.addErrback(on_transport_fail)

        return connect_defer

    def _open_channel(self, run):
        assert self.connection
        assert self.run_states[run.id].state < RUN_STATE_RUNNING

        self.run_states[run.id].state = RUN_STATE_STARTING

        chan = ssh.ExecChannel(conn=self.connection)

        chan.addOutputCallback(run.write_stdout)
        chan.addErrorCallback(run.write_stderr)
        chan.addEndCallback(run.done)

        chan.command = run.command
        chan.start_defer = defer.Deferred()
        chan.start_defer.addCallback(self._run_started, run)
        chan.start_defer.addErrback(self._run_start_error, run)

        chan.exit_defer = defer.Deferred()
        chan.exit_defer.addCallback(self._channel_complete, run)
        chan.exit_defer.addErrback(self._channel_complete_unknown, run)

        twistedutils.defer_timeout(chan.start_defer, RUN_START_TIMEOUT)

        self.run_states[run.id].channel = chan
        self.connection.openChannel(chan)

    def _channel_complete(self, channel, run):
        """Callback once our channel has completed it's operation

        This is how we let our run know that we succeeded or failed.
        """
        log.info("Run %s has completed with %r", run.id, channel.exit_status)
        if run.id not in self.run_states:
            log.warning("Run %s no longer tracked", run.id)
            return

        assert self.run_states[run.id].state < RUN_STATE_COMPLETE

        self.run_states[run.id].state = RUN_STATE_COMPLETE
        cb = self.run_states[run.id].deferred.callback
        self._cleanup(run)

        run.exited(channel.exit_status)
        cb(channel.exit_status)

    def _channel_complete_unknown(self, result, run):
        """Channel has closed on a running process without a proper exit

        We don't actually know if the run succeeded
        """
        log.error("Failure waiting on channel completion: %s", str(result))
        self._fail_run(run, failure.Failure(exc_value=ResultError()))

    def _run_started(self, channel, run):
        """Our run is actually a running process now, update the state"""
        log.info("Run %s started for %s", run.id, self.hostname)
        channel.start_defer = None
        assert self.run_states[run.id].state == RUN_STATE_STARTING
        self.run_states[run.id].state = RUN_STATE_RUNNING

        run.started()

    def _run_start_error(self, result, run):
        """We failed to even run the command due to communication difficulties

        Once all the runs have closed out we can try to reconnect.
        """
        log.error("Error running %s, disconnecting from %s: %s",
                  run.id, self.hostname, str(result))

        # We clear out the deferred that likely called us because there are
        # actually more than one error paths because of user timeouts.
        self.run_states[run.id].channel.start_defer = None

        self._fail_run(run, failure.Failure(
            exc_value=ConnectError("Connection to %s failed" % self.hostname)))

        # We want to hard hangup on this connection. It could theoretically
        # come back thanks to the magic of TCP, but something is up, best to
        # fail right now then limp along for and unknown amount of time.
        #self.connection.transport.connectionLost(failure.Failure())

    def repr_data(self):
        """Returns a dict which is an external view of this object."""
        return {
            'name':             self.name,
            'hostname':         self.hostname
        }

    def __str__(self):
        return "Node:%s" % self.hostname


"""
Parse and validate scheduler configuration and return immutable structures.
"""
import calendar
from collections import namedtuple
import datetime
import re

from tron.config import ConfigError


ConfigDailyScheduler = namedtuple(
    'ConfigDailyScheduler',
    ['ordinals', 'weekdays', 'monthdays', 'months', 'timestr']
)


ConfigConstantScheduler = namedtuple('ConfigConstantScheduler', [])


ConfigIntervalScheduler = namedtuple(
    'ConfigIntervalScheduler', [
        'timedelta',    # datetime.timedelta
    ])

class ScheduleParseError(Exception):
    pass



def valid_schedule(path, schedule):
    if isinstance(schedule, basestring):
        schedule = schedule.strip()
        scheduler_args = schedule.split()
        scheduler_name = scheduler_args.pop(0).lower()

        if schedule == 'constant':
            return ConfigConstantScheduler()
        elif scheduler_name == 'daily':
            return valid_daily_scheduler(*scheduler_args)
        elif scheduler_name == 'interval':
            return valid_interval_scheduler(' '.join(scheduler_args))
        else:
            return parse_daily_expression(schedule)

    if 'interval' in schedule:
        return valid_interval_scheduler(**schedule)
    elif 'start_time' in schedule or 'days' in schedule:
        return valid_daily_scheduler(**schedule)
    else:
        raise ConfigError("Unknown scheduler at %s: %s" % (path, schedule))


def valid_daily_scheduler(start_time=None, days=None):
    """Old style, will be converted to DailyScheduler with a compatibility
    function

    schedule:
        start_time: "07:00:00"
        days: "MWF"
    """

    err_msg = ("Start time must be in string format HH:MM[:SS]. Seconds"
               " are ignored but parsed so as to be backward-compatible."
               " You said: %s")

    if start_time is None:
        hms = ['00', '00']
    else:
        if not isinstance(start_time, basestring):
            raise ConfigError(err_msg % start_time)

        # make sure at least hours and minutes are specified
        hms = start_time.strip().split(':')

        if len(hms) < 2:
            raise ConfigError(err_msg % start_time)

    weekdays = set(CONVERT_DAYS_INT[d] for d in days or 'MTWRFSU')
    if weekdays == set([0, 1, 2, 3, 4, 5, 6]):
        days_str = 'day'
    else:
        # incoming string is MTWRF, we want M,T,W,R,F for the parser
        days_str = ','.join(days)

    return parse_daily_expression(
        'every %s of month at %s:%s' % (days_str, hms[0], hms[1])
    )


def valid_interval_scheduler(interval):
    # remove spaces
    interval = ''.join(interval.split())

    # Shortcut values for intervals
    TIME_INTERVAL_SHORTCUTS = {
        'hourly': dict(hours=1),
    }

    # Translations from possible configuration units to the argument to
    # datetime.timedelta
    TIME_INTERVAL_UNITS = {
        'months': ['mo', 'month', 'months'],
        'days': ['d', 'day', 'days'],
        'hours': ['h', 'hr', 'hrs', 'hour', 'hours'],
        'minutes': ['m', 'min', 'mins', 'minute', 'minutes'],
        'seconds': ['s', 'sec', 'secs', 'second', 'seconds']
    }

    if interval in TIME_INTERVAL_SHORTCUTS:
        kwargs = TIME_INTERVAL_SHORTCUTS[interval]
    else:
        # Split digits and characters into tokens
        interval_re = re.compile(r"\d+|[a-zA-Z]+")
        interval_tokens = interval_re.findall(interval)
        if len(interval_tokens) != 2:
            raise ConfigError("Invalid interval specification: %s", interval)

        value, units = interval_tokens

        kwargs = {}
        for key, unit_set in TIME_INTERVAL_UNITS.iteritems():
            if units in unit_set:
                kwargs[key] = int(value)
                break
        else:
            raise ConfigError("Invalid interval specification: %s", interval)

    return ConfigIntervalScheduler(timedelta=datetime.timedelta(**kwargs))


def day_canonicalization_map():
    """Build a map of weekday synonym to int index 0-6 inclusive."""
    canon_map = dict()

    # 7-element lists with weekday names in order
    weekday_lists = (calendar.day_name,
                     calendar.day_abbr,
                     ('m', 't', 'w', 'r', 'f', 's', 'u'),
                     ('mo', 'tu', 'we', 'th', 'fr', 'sa', 'su'))
    for day_list in weekday_lists:
        for day_name_synonym, day_index in zip(day_list, range(7)):
            canon_map[day_name_synonym] = day_index
            canon_map[day_name_synonym.lower()] = day_index
            canon_map[day_name_synonym.upper()] = day_index
    return canon_map

# Canonicalize weekday names to integer indices
CONVERT_DAYS_INT = day_canonicalization_map()   # day name/abbrev => {0123456}


def month_canonicalization_map():
    """Build a map of month synonym to int index 0-11 inclusive."""
    canon_map = dict()

    # calendar stores month data with a useless element in front. cut it off.
    monthname_lists = (calendar.month_name[1:], calendar.month_abbr[1:])
    for month_list in monthname_lists:
        for key, value in zip(month_list, range(1, 13)):
            canon_map[key] = value
            canon_map[key.lower()] = value
    return canon_map


# Canonicalize month names to integer indices
# month name/abbrev => {0 <= k <= 11}
CONVERT_MONTHS = month_canonicalization_map()


def daily_schedule_parser_re():
    """Build a regular expression that matches this:

        ("every"|ordinal) (day) ["of|in" (monthspec)] (["at"] HH:MM)

    ordinal   - comma-separated list of "1st" and so forth
    days      - comma-separated list of days of the week (for example,
                "mon", "tuesday", with both short and long forms being
                accepted); "every day" is equivalent to
                "every mon,tue,wed,thu,fri,sat,sun"
    monthspec - comma-separated list of month names (for example,
                "jan", "march", "sep"). If omitted, implies every month.
                You can also say "month" to mean every month, as in
                "1,8th,15,22nd of month 09:00".
    HH:MM     - time of day in 24 hour time.

    This is a slightly more permissive version of Google App Engine's schedule
    parser, documented here:
    http://code.google.com/appengine/docs/python/config/cron.html#The_Schedule_Format
    """

    # m|mon|monday|...|day
    DAY_VALUES = '|'.join(CONVERT_DAYS_INT.keys() + ['day'])

    # jan|january|...|month
    MONTH_VALUES = '|'.join(CONVERT_MONTHS.keys() + ['month'])

    DATE_SUFFIXES = 'st|nd|rd|th'

    # every|1st|2nd|3rd (also would accept 3nd, 1rd, 4st)
    MONTH_DAYS_EXPR = '(?P<month_days>every|((\d+(%s),?)+))?' % DATE_SUFFIXES
    DAYS_EXPR = r'((?P<days>((%s),?)+))?' % DAY_VALUES
    MONTHS_EXPR = r'((in|of)\s+(?P<months>((%s),?)+))?' % MONTH_VALUES

    # [at] 00:00
    TIME_EXPR = r'((at\s+)?(?P<time>\d\d:\d\d))?'

    DAILY_SCHEDULE_EXPR = ''.join([
        r'^',
        MONTH_DAYS_EXPR, r'\s*',
        DAYS_EXPR, r'\s*',
        MONTHS_EXPR, r'\s*',
         TIME_EXPR, r'\s*',
        r'$'
    ])
    return re.compile(DAILY_SCHEDULE_EXPR)


# Matches expressions of the form
# ``("every"|ordinal) (days) ["of|in" (monthspec)] (["at"] HH:MM)``.
# See :py:func:`daily_schedule_parser_re` for details.
DAILY_SCHEDULE_RE = daily_schedule_parser_re()


def _parse_number(day):
    return int(''.join(c for c in day if c.isdigit()))

def parse_daily_expression(expression):
    """Given an expression of the form in the docstring of
    daily_schedule_parser_re(), return the parsed values in a
    ConfigDailyScheduler
    """
    m = DAILY_SCHEDULE_RE.match(expression.lower())
    if not m:
        raise ScheduleParseError('Expression %r is not a valid scheduler'
                                 ' expression.' % expression)

    timestr = m.group('time')
    if timestr is None:
        timestr = '00:00'

    if m.group('days') in (None, 'day'):
        weekdays = None
    else:
        weekdays = set(CONVERT_DAYS_INT[d] for d in m.group('days').split(','))

    monthdays = None
    ordinals = None

    if m.group('month_days') != 'every':
        values = set(_parse_number(n)
                     for n in m.group('month_days').split(','))
        if weekdays is None:
            monthdays = values
        else:
            ordinals = values

    if m.group('months') in (None, 'month'):
        months = None
    else:
        months = set(CONVERT_MONTHS[mo] for mo in m.group('months').split(','))

    return ConfigDailyScheduler(
        ordinals=ordinals,
        weekdays=weekdays,
        monthdays=monthdays,
        months=months,
        timestr=timestr,
    )